from __future__ import annotations

import abc
import enum
import os.path
import warnings
import threading

from typing import Union, Dict, Any, AnyStr, Literal

from .keyring import Key
from .minivault import MiniVault
from .vaultstructs import VaultStructBase


class ModeProperties(dict):
    def __setattr__(self, key, value):
        if hasattr(self, "initialized") and self.initialized:
            raise AttributeError("ModeProperties are immutable")
        super(ModeProperties, self).__setattr__(key, value)
        super(ModeProperties, self).__setitem__(key, value)

    def __init__(self, read_only: bool, live_update: bool, create: bool, load: bool, write: bool):
        self.initialized = False
        super().__init__()
        self.read_only = read_only
        self.live_update = live_update
        self.create = create
        self.load = load
        self.write = write
        self.initialized = True


class ResourceModes(enum.Enum):
    READ = "r"                     # Read from existing resource (default)
    WRITE = "w"                    # Create new resource and ignore existing resource and write to it
    APPEND = "a"                   # Create a new resource if none exist, otherwise read from and write to existing resource
    READ_W_LIVE_UPDATE = "r+"      # Read from existing resource and perform live-update
    WRITE_W_LIVE_UPDATE = "w+"     # Create new resource and ignore existing resource and write to it, and perform live-update
    APPEND_W_LIVE_UPDATE = "a+"    # Create a new resource if none exist, otherwise read from and write to existing resource, and perform live-update


class BaseResource(abc.ABC):

    POSSIBLE_MODES = {
        ResourceModes.READ.value,
        ResourceModes.WRITE.value,
        ResourceModes.APPEND.value,
        ResourceModes.READ_W_LIVE_UPDATE.value,
        ResourceModes.WRITE_W_LIVE_UPDATE.value,
        ResourceModes.APPEND_W_LIVE_UPDATE.value,
    }

    MODE_MAPPING: Dict[str, ModeProperties] = {
        ResourceModes.READ.value:                  ModeProperties(read_only=True,  live_update=False, create=False, load=True,  write=False),
        ResourceModes.WRITE.value:                 ModeProperties(read_only=False, live_update=False, create=True,  load=False, write=True),
        ResourceModes.APPEND.value:                ModeProperties(read_only=False, live_update=False, create=True,  load=True,  write=True),
        ResourceModes.READ_W_LIVE_UPDATE.value:    ModeProperties(read_only=True,  live_update=True,  create=False, load=True,  write=False),
        ResourceModes.WRITE_W_LIVE_UPDATE.value:   ModeProperties(read_only=False, live_update=True,  create=True,  load=False, write=True),
        ResourceModes.APPEND_W_LIVE_UPDATE.value:  ModeProperties(read_only=False, live_update=True,  create=True,  load=True,  write=True)
    }

    def __init__(self, path: Union[AnyStr, Any], mode: Union[Literal["r", "w", "a", "r+", "w+", "a+"], ResourceModes] = "r"):
        """
        Creates an object that can be used to read and write to a resource.

        :param path: This is the path to the resource. It can be essentially anything that identifies it. Any resource has a path to where it is located.
        :param mode: Sets the mode of the resource. The mode can be one of the following: 'r', 'w', 'a', 'r+', 'w+', 'a+'.
        r: Read from existing resource (default)
        w: Create new resource and ignore existing resource and write to it
        a: Create a new resource if none exist, otherwise read from and write to existing resource
        r+: Read from existing resource and perform live-update
        w+: Create new resource and ignore existing resource and write to it, and perform live-update
        a+: Create a new resource if none exist, otherwise read from and write to existing resource, and perform live-update

        """
        self.lock = threading.Lock()

        if isinstance(mode, ResourceModes):
            mode = mode.value
        self.mode = mode
        if self.mode not in self.POSSIBLE_MODES:
            raise ValueError(f"Invalid mode: {self.mode}. Must be one of the following: {self.POSSIBLE_MODES}")

        self.raw_path = os.path.expanduser(os.path.expandvars(path))

        self.mode_properties: ModeProperties = self.MODE_MAPPING[self.mode]
        self.last_known_state = None
        self.cached_state = None
        if self.mode_properties.load and not self.mode_properties.create and not self.exists() and not self.mode_properties.live_update:
            raise ResourceNotFoundError(f"Resource not found at: {self.raw_path} (mode is {self.mode})", self)

    def resource_has_changed(self):
        """Returns a bool that says if the resource has changed since the last time it was read."""
        self.last_known_state = self.state
        return self.cached_state != self.last_known_state

    def update_state(self, fetch_state=True):
        """Updates the state by fetching the current state."""
        if fetch_state:
            self.last_known_state = self.state
        self.cached_state = self.last_known_state

    def create_mv(self, **keys: Key) -> MiniVault:
        f"""Creates a {MiniVault}-object from a file by loading the vault from the file using the keyring."""
        from varvault import concurrent_execution

        vault_file_data = self.read()

        assert isinstance(vault_file_data, dict), f"'vault_file_data' from the filehandler is not a dict: {vault_file_data}"

        # Get the keys from the keyring as a list.
        return_vault_data = dict()

        async def build(key_in_file: str):
            if key_in_file not in keys:
                return
            key: Key = keys[key_in_file]
            if key.valid_type and issubclass(key.valid_type, VaultStructBase):
                return_vault_data[key] = key.valid_type.create(key_in_file, vault_file_data[key_in_file])
            else:
                if key.can_be_none and vault_file_data[key_in_file] is None:
                    return_vault_data[key] = None
                else:
                    assert key.valid_type is None or isinstance(vault_file_data[key_in_file], key.valid_type), f"Key type missmatch ({key}; Valid type {key.valid_type}, actual type: {type(vault_file_data[key_in_file])}"
                    return_vault_data[key] = vault_file_data[key_in_file]

        concurrent_execution(build, vault_file_data.keys())

        return MiniVault(**return_vault_data)

    @property
    @abc.abstractmethod
    def resource(self) -> Any:
        """Meant to return the resource that stores the vault in some database such as a file."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def state(self):
        """Meant to return the state of the resource, such as a hash of the resource."""
        raise NotImplementedError()

    @abc.abstractmethod
    def create(self):
        """Meant to create the resource to store the vault in a database."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def path(self) -> Union[AnyStr, Any]:
        """Meant to return the path to the database that stores the vault."""
        raise NotImplementedError()

    @abc.abstractmethod
    def writable(self, obj: Dict) -> bool:
        """Meant to return a bool that says if a given key-value pair in a dict can be successfully written to the database."""
        raise NotImplementedError()

    @abc.abstractmethod
    def exists(self) -> bool:
        """Meant to return a bool which says if the resource exists or not"""
        raise NotImplementedError()

    # ================================================================================================================
    # Write
    # ================================================================================================================
    def write(self, vault: dict) -> None:
        f"""Writes the vault to the database by calling the implemented '{self.do_write}' method. Not meant to be overridden."""
        if not vault:
            # No point writing an empty dict and it's not the job of this method to create the file
            return

        if self.mode_properties.read_only:
            warnings.warn("Tried to write to a resource defined as read-only. This is not permitted by varvault.")
            return

        if not self.resource:
            self.create()
        try:
            with self.lock:
                self.do_write(vault)
        except Exception as e:
            raise ResourceNotFoundError(f"Failed to write to the resource: {e}", self)
        self.update_state()

    @abc.abstractmethod
    def do_write(self, vault: dict) -> None:
        """
        A function to write a vault to a file/database. Varvault will call this function internally.
        The class that implements this abstract method has to write a dict to a resource.

        Example:

        ``json.dump(vault, open(self.path, "w"), indent=2)``

        :param vault: The vault to write to the file.
        :return: None. Varvault will not use the return value from this function
        """
        raise NotImplementedError()

    # ================================================================================================================
    # Read
    # ================================================================================================================
    def read(self) -> Dict:
        f"""Reads the vault from the database by calling the implemented '{self.do_read}' method. Not meant to be overridden."""
        if not self.resource:
            self.create()
        with self.lock:
            if self.exists():
                try:
                    data = self.do_read()
                    self.update_state()
                    return data
                except Exception as e:
                    raise ResourceNotFoundError(f"Failed to read from the resource (mode is {self.mode}): {e}", self)
            if self.mode_properties.live_update:
                return {}
            raise ResourceNotFoundError(f"Resource not found at: {self.raw_path} (mode is {self.mode})", self)

    @abc.abstractmethod
    def do_read(self) -> Dict:
        """
        A function to read data from a file/database. Varvault will call this function internally.
        The class that implements this abstractmethod has to read data from a file and then return it.

        Example:

        ``return json.load(open(self.path))``

        :return: A dict describing the vault from the resource.
        """
        raise NotImplementedError()

    def __str__(self):
        return f"resource={self.resource}; path={self.path}; live_update={self.mode_properties.live_update}; vault_is_read_only={self.mode_properties.read_only}"


class ResourceNotFoundError(FileNotFoundError):
    def __init__(self, msg, resource: BaseResource):
        super(ResourceNotFoundError, self).__init__(msg)
        self.msg = msg
        self.resource = resource

    def __repr__(self):
        return f"ResourceNotFoundError: {self.msg}; {self.resource.mode_properties}"

    def __str__(self):
        return f"{self.msg} - {self.resource.mode_properties}"



import json
import os.path
import re
from typing import Callable

import pytest
import tempfile

from _pytest.recwarn import warns

from commons import *

vault_file_new = f"{DIR}/new-vault.json"
vault_file_new_secondary = f"{DIR}/new-vault-secondary.json"
existing_vault = f"{DIR}/existing-vault.json"
faulty_existing_vault = f"{DIR}/faulty-existing-vault.json"
faulty_vault_key_missmatch = f"{DIR}/faulty-vault-key-missmatch.json"


class TestVault:

    @classmethod
    def setup_class(cls):
        tempfile.tempdir = "/tmp" if sys.platform == "darwin" or sys.platform == "linux" else tempfile.gettempdir()

    def setup_method(self):
        try:
            os.remove(vault_file_new)
        except:
            pass
        try:
            os.remove(vault_file_new_secondary)
        except:
            pass

    def test_create_new_vault(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=Keyring.key_valid_type_is_str)
        def _set_valid():
            return "valid-key"

        _set_valid()
        assert vault.get(Keyring.key_valid_type_is_str) == "valid-key"

        @vault.manual(varvault.Flags.permit_modifications, input=Keyring.key_valid_type_is_str, output=Keyring.key_valid_type_is_str)
        def _replace(key_valid_type_is_str):
            return f"{key_valid_type_is_str}-replaced"

        _replace()

        @vault.manual(output=Keyring.key_valid_type_is_int)
        def _set_invalid():
            return "invalid-key; must be int"

        try:
            _set_invalid()
            pytest.fail(f"Somehow managed to set an invalid value to key {Keyring.key_valid_type_is_int} (valid type: {Keyring.key_valid_type_is_int.valid_type})")
        except Exception as e:
            assert "Key 'key_valid_type_is_int' requires type to be '<class 'int'>'" in str(e), f"Unexpected error: {e}"
            logger.info(f"Expected error received; test passed")
            assert Keyring.key_valid_type_is_int not in vault

    def test_no_keys(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        called = False
        vault.manual()

        def _set_valid():
            nonlocal called
            called = True
            return "valid-key"

        _set_valid()
        assert called

    def test_put(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        mv = varvault.MiniVault({Keyring.key_valid_type_is_str: "value", Keyring.key_valid_type_is_int: 1})
        vault._put(mv)
        assert Keyring.key_valid_type_is_str in vault
        assert Keyring.key_valid_type_is_int in vault

        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        vault._put(Keyring.key_valid_type_is_str, "value")
        assert Keyring.key_valid_type_is_str in vault
        assert Keyring.key_valid_type_is_int not in vault

    def test_create_create_vault(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(existing_vault, mode="r"))
        logger.info(vault)
        assert Keyring.key_valid_type_is_str in vault
        assert Keyring.key_valid_type_is_int in vault
        assert vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert vault.get(Keyring.key_valid_type_is_int) == 1
        d = json.load(open(existing_vault))
        assert Keyring.key_valid_type_is_str in d and Keyring.key_valid_type_is_int in d, "It appears that loading from the vault file has cleared the vault file unintentionally. This is very bad"

    def test_load_from_one_write_to_another(self):
        existing = varvault.JsonResource(existing_vault, "r").create_mv(**Keyring.get_keys())
        logger.info(f"Existing vault: {existing}")
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        vault.insert_minivault(existing)

        @vault.manual(varvault.Flags.permit_modifications, input=Keyring.key_valid_type_is_str, output=Keyring.key_valid_type_is_str)
        def mod(**kwargs):
            key_valid_type_is_str = kwargs.get(Keyring.key_valid_type_is_str)
            assert key_valid_type_is_str == "valid"
            modded = f"modded"
            return modded
        mod()

        assert vault.get(Keyring.key_valid_type_is_str) == "modded"
        assert json.load(open(vault_file_new)).get(Keyring.key_valid_type_is_str) == "modded", f"The value for {Keyring.key_valid_type_is_str} in vault_filename_to is not the expected"
        assert json.load(open(existing_vault)).get(Keyring.key_valid_type_is_str) == "valid", f"The value for {Keyring.key_valid_type_is_str} in vault_filename_from has changed. This is very bad"

    def test_create_create_vault_no_valid_type_in_key(self):
        class KeyringTemp(varvault.Keyring):
            key_valid_type_is_str = varvault.Key("key_valid_type_is_str")
            key_valid_type_is_int = varvault.Key("key_valid_type_is_int")

        vault = varvault.create(keyring=KeyringTemp, resource=varvault.JsonResource(existing_vault, mode="r"))

    def test_permit_modifications(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        vault.insert(Keyring.key_valid_type_is_str, "valid")
        try:
            @vault.manual(output=Keyring.key_valid_type_is_str)
            def _set():
                return "new-value-that-should-not-go-in"
            _set()

            pytest.fail("Managed to set a new value to an existing key while modifications are not permitted")
        except Exception as e:
            logger.info(f"Expected error received; test passed")
            assert vault.get(Keyring.key_valid_type_is_str) == "valid", f"Value for {Keyring.key_valid_type_is_str} is not what it should be"

        @vault.manual(varvault.Flags.permit_modifications, output=Keyring.key_valid_type_is_str)
        def _set():
            return "new-modified-value"
        _set()

        assert vault.get(Keyring.key_valid_type_is_str) == "new-modified-value", f"Value for {Keyring.key_valid_type_is_str} is not what it should be"

        new_vault = varvault.create(varvault.Flags.permit_modifications, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @new_vault.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "new-modified-value-gen-2"
        _set()

        assert new_vault.get(Keyring.key_valid_type_is_str) == "new-modified-value-gen-2", f"Value for {Keyring.key_valid_type_is_str} is not what it should be"

    def test_create_readonly_vault(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(existing_vault, mode="r"))
        try:
            vault.insert(Keyring.key_valid_type_is_int, 1)
            pytest.fail("Insert: Somehow managed to insert a value into a vault that is supposed to be read-only")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

        try:
            @vault.manual(output=Keyring.key_valid_type_is_int)
            def _set():
                return 1
            _set()

            pytest.fail("Vaulter: Somehow managed to insert a value into a vault that is supposed to be read-only")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

    def test_read_only_key_not_in_keyring(self):
        with warns(UserWarning, match=".*is not defined in the keyring.*"):
            json.dump({Keyring.key_valid_type_is_str: "valid", Keyring.key_valid_type_is_int: 1, "temp": "this-should-not-be-in-the-vault"}, open(vault_file_new, "w"))
            vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="r"))
            assert varvault.Key("temp") not in vault, "Vault contains a key that should not be in the vault since it doesn't exist in the keyring"
            assert Keyring.key_valid_type_is_str in vault
            assert Keyring.key_valid_type_is_int in vault

    def test_insert_nonexistent_key(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        temp_key = varvault.Key("temp_key")
        try:
            vault.insert(temp_key, "this-should-not-go-in")
            pytest.fail("Somehow managed to insert a non-existent key into a vault that should not permit this")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

    # @pytest.mark.filterwarnings("ignore:.*Keys were found in the resource.*", "ignore:.*is not defined in the keyring.*")
    def test_create_from_faulty_vault(self):
        with warns(UserWarning, match=".*is not defined in the keyring.*|.*Keys were found in the resource.*"):
            this_key_doesnt_exist_in_keyring = varvault.Key("this_key_doesnt_exist_in_keyring", valid_type=str)
            try:
                vault = varvault.create(keyring=Keyring, logger=logger,
                                        resource=varvault.JsonResource(faulty_existing_vault, mode="r"))
                pytest.fail("Managed to create a vault from a file that should be faulty")
            except Exception as e:
                logger.info(f"Expected error received; test passed: {e}")

            try:
                vault = varvault.create(keyring=Keyring,
                                        logger=logger,
                                        resource=varvault.JsonResource(faulty_vault_key_missmatch, mode="a"))
                pytest.fail("Managed to create a vault from a file with a key not in keyring, and ignore_keys_not_in_keyring is False and the resource is not read-only (it's in append mode)")
            except Exception as e:
                logger.info(f"Expected error received; test passed: {e}")

            vault = varvault.create(varvault.Flags.ignore_keys_not_in_keyring, keyring=Keyring, name="from-vault",
                                    logger=logger,
                                    resource=varvault.JsonResource(faulty_vault_key_missmatch, mode="r"))
            assert this_key_doesnt_exist_in_keyring not in vault, f"Key {this_key_doesnt_exist_in_keyring} was found in the vault when it shouldn't be"

            vault = varvault.create(keyring=Keyring,
                                    resource=varvault.JsonResource(faulty_vault_key_missmatch, mode="r"),
                                    logger=logger,
                                    this_key_doesnt_exist_in_keyring=this_key_doesnt_exist_in_keyring)
            assert this_key_doesnt_exist_in_keyring in vault, f"Key {this_key_doesnt_exist_in_keyring} was not found in the vault when it should be added as an extra key"

    def test_insert_type_validation(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        try:
            vault.insert(Keyring.key_valid_type_is_int, "this-should-not-work")
            assert False, "Somehow managed to insert a value for a key that should not work"
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

    def test_clean_output_keys(self):
        vault_new = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault_new.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()

        assert vault_new.get(Keyring.key_valid_type_is_str) == "valid"

        @vault_new.manual(varvault.Flags.clean_output_keys, output=Keyring.key_valid_type_is_str)
        def _clean():
            return

        _clean()
        assert Keyring.key_valid_type_is_str in vault_new, f"No {Keyring.key_valid_type_is_str} in vault"
        assert vault_new.get(Keyring.key_valid_type_is_str) == "", f"Key {Keyring.key_valid_type_is_str} is not an empty string; {vault_new.get(Keyring.key_valid_type_is_str)}"

    def test_extra_keys(self):
        extra_key1 = varvault.Key("extra_key1", valid_type=dict)
        vault_new = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"), extra_key1=varvault.Key("extra_key1", valid_type=dict))

        @vault_new.manual(output=extra_key1)
        def _set_invalid():
            return [1, 2, 3]
        try:
            _set_invalid()
            pytest.fail("Unexpectedly managed to set an invalid value to an extra key")
        except Exception as e:
            assert "Key 'extra_key1' requires type to be '<class 'dict'>'" in str(e), f"Unexpected error message: {e}"
            logger.info(f"Expected error received; test passed: {e}")

        @vault_new.manual(output=extra_key1)
        def _set_valid():
            return {"a": 1, "b": 2, "c": 3}
        _set_valid()

        @vault_new.manual(varvault.Flags.clean_output_keys, output=extra_key1)
        def _clean():
            return
        _clean()

        assert vault_new.get(extra_key1) == {}

    def test_return_tuple_is_single_item(self):
        tuple_item = varvault.Key("tuple_item", valid_type=tuple)
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"), tuple_item=tuple_item)

        @vault.manual(varvault.Flags.return_tuple_is_single_item, output=tuple_item)
        def _set():
            return 1, 2, 3
        _set()

        assert tuple_item in vault, f"Flag: No {tuple_item} found in vault"
        assert vault.get(tuple_item) == (1, 2, 3), "Flag: missmatch"

        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"), tuple_item=tuple_item)

        @vault.manual(output=tuple_item)
        def _set():
            return 1, 2, 3
        _set()

        assert tuple_item in vault, f"No flag: No {tuple_item} found in vault"
        assert vault.get(tuple_item) == (1, 2, 3), "No flag: Missmatch"

    def test_split_output_keys(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        vault_secondary = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new_secondary, mode="w"))

        @vault.manual(varvault.Flags.split_output_keys, output=Keyring.key_valid_type_is_str)
        @vault_secondary.manual(varvault.Flags.split_output_keys, output=Keyring.key_valid_type_is_int)
        def _set():
            return varvault.MiniVault({Keyring.key_valid_type_is_str: "valid", Keyring.key_valid_type_is_int: 1})
        _set()

        assert Keyring.key_valid_type_is_str in vault and Keyring.key_valid_type_is_int not in vault
        assert Keyring.key_valid_type_is_int in vault_secondary and Keyring.key_valid_type_is_str not in vault_secondary

        assert vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert vault_secondary.get(Keyring.key_valid_type_is_int) == 1

    def test_return_key_can_be_missing(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def _set_failed():
            return "valid"

        try:
            # Should fail saying that number of returned items do not match the number of keys
            _set_failed()
            pytest.fail("Managed to set a single variable to two keys or something")
        except Exception:
            logger.info("Expected error received; test passed")

        @vault.manual(varvault.Flags.output_key_can_be_missing, output=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def _set_failed_again():
            return "valid"

        try:
            _set_failed_again()
            pytest.fail(f"Managed to set a single variable when {varvault.Flags.output_key_can_be_missing} is defined; "
                        f"Should have failed saying return var must be of type {varvault.MiniVault}")
        except Exception:
            logger.info("Expected error received; test passed")

        @vault.manual(varvault.Flags.output_key_can_be_missing, output=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def _set_working():
            return varvault.MiniVault({Keyring.key_valid_type_is_str: "valid"})

        _set_working()
        assert Keyring.key_valid_type_is_str in vault
        assert Keyring.key_valid_type_is_int not in vault
        assert vault.get(Keyring.key_valid_type_is_str) == "valid"

    def test_validate_types_in_minivault_return_values(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def _set_failed():
            return varvault.MiniVault({Keyring.key_valid_type_is_str: 1, Keyring.key_valid_type_is_int: "invalid"})
        try:
            _set_failed()
            pytest.fail("Managed to set invalid values to the vault by returning them in a minivault.")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

    def test_validator_decorator(self):
        try:
            @varvault.validator(function_asserts=True, function_returns_bool=True)
            def invalid_decorator_use(value):
                pass
            pytest.fail("Managed to register a validator function that should not be possible to register. Both 'function_asserts' and 'function_returns_bool' should not be possible to set at the same time")
        except SyntaxError as e:
            logger.info(f"Expected error received; test passed: {e}")

        try:
            @varvault.validator()
            def invalid_decorator_use(value, this_arg_must_not_exist):
                pass
            pytest.fail("Managed to register a validator function that should not be possible to register. Managed to register a validator function with more than 1 positional arg called 'keyvalue'")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

        try:
            @varvault.validator(function_asserts=True)
            def invalid_decorator_use(value):
                pass
            pytest.fail("Managed to register a validator function that should not be possible to register. The function needs to contain a call 'assert' in the source if 'function_asserts' is set to True")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

        try:
            @varvault.validator(function_returns_bool=True)
            def invalid_decorator_use(value) -> bool:
                pass
            pytest.fail("Managed to register a validator function that should not be possible to register. The function needs to return something in the source if 'function_returns_bool' is set to True")
        except Exception as e:
            logger.info(f"Expected error received; test passed: {e}")

        @varvault.validator(function_asserts=True, skip_source_assertions=True)
        def valid_decorator_use_1(value):
            pass

        @varvault.validator(function_returns_bool=True, skip_source_assertions=True)
        def valid_decorator_use_2(value) -> bool:
            pass

    def test_key_validation_function(self):
        # This will validate the function and assign some attributes to it that varvault will use when validating the value for the key
        @varvault.validator(function_returns_bool=True)
        def must_be_even(value: int) -> bool:
            return (value % 2) == 0

        with warns(SyntaxWarning, match="If you define the validator-function to say that it returns a bool"):
            @varvault.validator(function_returns_bool=True)
            def returns_no_bool(value) -> int:
                return 1

        @varvault.validator(function_asserts=True)
        def cannot_be_negative(value: int):
            assert value >= 0, "Value cannot be negative"

        @varvault.validator()
        def no_dashes(value: str):
            assert "-" not in value

        class KeyringKeyValidationFunction(varvault.Keyring):
            int_must_be_even_number = varvault.Key("int_must_be_even_number", valid_type=int, validators=must_be_even)
            int_cannot_be_negative = varvault.Key("int_cannot_be_negative", valid_type=int, validators=cannot_be_negative)
            validator_returns_no_bool = varvault.Key("validator_returns_no_bool", valid_type=int, validators=returns_no_bool)
            no_dashes_in_str = varvault.Key("no_dashes_in_str", valid_type=str, validators=no_dashes)

        vault = varvault.create(keyring=KeyringKeyValidationFunction, resource=varvault.JsonResource(vault_file_new, mode="w"))

        with pytest.raises(varvault.ValidatorException) as e:
            vault.insert(KeyringKeyValidationFunction.int_must_be_even_number, 1)
        assert "must_be_even" in str(e.value)

        @vault.manual(output=KeyringKeyValidationFunction.int_must_be_even_number)
        def set_failed():
            return 5
        with pytest.raises(varvault.ValidatorException) as e:
            set_failed()
        assert "must_be_even" in str(e.value)

        with pytest.raises(varvault.ValidatorException) as e:
            vault.insert(KeyringKeyValidationFunction.int_cannot_be_negative, -2)
        assert "Value cannot be negative" in str(e.value)

        vault.insert(KeyringKeyValidationFunction.int_must_be_even_number, 2)
        assert KeyringKeyValidationFunction.int_must_be_even_number in vault and vault.get(KeyringKeyValidationFunction.int_must_be_even_number) == 2

        with pytest.raises(varvault.ValidatorException) as e:
            vault.insert(KeyringKeyValidationFunction.validator_returns_no_bool, 1)
        assert "The return value is of type" in str(e.value)

        @vault.manual(varvault.Flags.permit_modifications, output=KeyringKeyValidationFunction.int_must_be_even_number)
        def set():
            return 4
        set()

        assert KeyringKeyValidationFunction.int_must_be_even_number in vault and vault.get(KeyringKeyValidationFunction.int_must_be_even_number) == 4

    def test_key_validation_error_message(self):
        @varvault.validator(function_asserts=True)
        def must_be_above_zero(value: int):
            assert value >= 0

        class KeyringTemp(varvault.Keyring):
            int_must_be_above_zero = varvault.Key("int_must_be_above_zero", valid_type=int, validators=must_be_above_zero)

        vault = varvault.create(keyring=KeyringTemp, name="vault")
        try:
            vault.insert(KeyringTemp.int_must_be_above_zero, -1)
            pytest.fail(f"Managed to set {KeyringTemp.int_must_be_above_zero} to a negative number via {vault.insert.__name__}. This should not be possible.")
        except Exception as e:
            assert KeyringTemp.int_must_be_above_zero in str(e), f"Expected the key {KeyringTemp.int_must_be_above_zero} to be mentioned in the error message, but it was not. Error message: {e}"
            logger.info(e)

    def test_add_minivault_function(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def insert():
            mv = varvault.MiniVault()
            mv.add(Keyring.key_valid_type_is_str, "valid")
            mv.add(Keyring.key_valid_type_is_int, 1)

            return mv

        insert()

        assert Keyring.key_valid_type_is_str in vault and vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert Keyring.key_valid_type_is_int in vault and vault.get(Keyring.key_valid_type_is_int) == 1

    def test_create_vault_file_in_non_existent_dir(self):
        temp_dir = f"{DIR}/temp-dir"
        vault_file = f"{temp_dir}/vault.json"

        try:
            os.remove(vault_file)
        except OSError:
            pass
        try:
            os.removedirs(temp_dir)
        except OSError:
            pass
        assert not os.path.exists(temp_dir), f"Dir {temp_dir} already exists. It's supposed to not exist before we create the vault to make sure varvault creates the required directories"
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(f"{temp_dir}/vault.json", mode="w"))
        vault.insert(Keyring.key_valid_type_is_str, "valid")
        vault.insert(Keyring.key_valid_type_is_int, 1)
        data = json.load(open(vault_file))
        assert Keyring.key_valid_type_is_str in data and Keyring.key_valid_type_is_int in data

    def test_get_with_default(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        vault.insert(Keyring.key_valid_type_is_str, "valid")
        assert Keyring.key_valid_type_is_int not in vault
        try:
            v = vault.get(Keyring.key_valid_type_is_int, default=1)
            pytest.fail(f"Managed to get value from vault with default set, without configuring {varvault.Flags.input_key_can_be_missing}: {v}")
        except Exception as e:
            pass

        v = vault.get(Keyring.key_valid_type_is_int, varvault.Flags.input_key_can_be_missing)
        assert v is None

        v = vault.get(Keyring.key_valid_type_is_int, varvault.Flags.input_key_can_be_missing, default=1)
        assert v == 1

        v = vault.get(Keyring.key_valid_type_is_int, varvault.Flags.input_key_can_be_missing) or 2
        assert v == 2

    def test_get_multiple_with_input_key_can_be_missing_flag(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        vault.insert(Keyring.key_valid_type_is_str, "valid")

        assert Keyring.key_valid_type_is_int not in vault

        @vault.manual(input=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def noflag(**kwargs):
            pass

        try:
            noflag()
            pytest.fail(f"We managed to get this far, which shouldn't be possible: {vault}")
        except:
            pass

        @vault.manual(varvault.Flags.input_key_can_be_missing, input=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def withflag(key_valid_type_is_str: str = None, key_valid_type_is_int: int = None):
            key_valid_type_is_int = key_valid_type_is_int or 1
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1, f"'key_valid_type_is_int' was not {None} when it came into the function: {key_valid_type_is_int}"

        withflag()

    def test_filehandler_exists(self):
        dir = "~/.varvault-test-dir"
        dir_expanded = os.path.expanduser(dir)
        file = "~/.varvault-test-dir/vault.json"
        file_expanded = os.path.expanduser(file)
        try:
            assert not os.path.exists(file_expanded), f"File {file_expanded} already exists, which means it may contain something that should not be overwritten."
            os.makedirs(dir_expanded)
            with open(file_expanded, "w") as f:
                f.write("{}")
            os.environ["DIR"] = "~/.varvault-test-dir"
            os.environ["VAULT_FILE"] = "vault.json"

            fh = varvault.JsonResource("$DIR/$VAULT_FILE")
            assert fh.exists()
        finally:
            try:
                os.remove(file_expanded)
            except OSError:
                pass
            try:
                os.removedirs(dir_expanded)
            except OSError:
                pass

        assert not os.path.exists(file_expanded), f"The file {file_expanded} still exists and it shouldn't. You may want to delete it manually"
        assert not os.path.exists(dir_expanded), f"The directory {dir_expanded} still exists and it shouldn't. You may want to delete it manually"
        logger.info("Done")

    def test_create_vault_produces_valid_json(self):
        assert not os.path.exists(vault_file_new)
        # Creating a vault from scratch should produce a valid but empty JSON file
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        assert json.load(open(vault_file_new)) == {}

    def test_input_keys_as_kw_vars_only(self):
        vault = varvault.create(varvault.Flags.use_signature_for_input_keys, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        vault.insert(Keyring.key_valid_type_is_str, "valid")
        vault.insert(Keyring.key_valid_type_is_int, 1)

        @vault.manual()
        def fn_pure_kw_only(*, key_valid_type_is_str=varvault.AssignedByVault, key_valid_type_is_int=varvault.AssignedByVault):
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1

        fn_pure_kw_only()

        @vault.manual()
        def fn_pure_kw_or_positional(key_valid_type_is_str=varvault.AssignedByVault, key_valid_type_is_int=varvault.AssignedByVault):
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1

        fn_pure_kw_or_positional()

        @vault.manual(input=Keyring.key_valid_type_is_str)
        def fn_mixed(key_valid_type_is_str=varvault.AssignedByVault, key_valid_type_is_int=varvault.AssignedByVault):
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1

        fn_mixed()

        @vault.manual()
        def fn_pure_with_args(a1, key_valid_type_is_str=varvault.AssignedByVault, key_valid_type_is_int=varvault.AssignedByVault):
            assert a1 == 3.14
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1

        fn_pure_with_args(3.14)

        @vault.manual(input=Keyring.key_valid_type_is_str)
        def fn_mixed_with_args(a1, key_valid_type_is_str=varvault.AssignedByVault, key_valid_type_is_int=varvault.AssignedByVault):
            assert a1 == 3.14
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1

        fn_mixed_with_args(3.14)

        try:
            @vault.manual()
            def fn_with_typo(key_valid_type_is_string=varvault.AssignedByVault, key_valid_type_is_integer=varvault.AssignedByVault):  # Note the typos in the key names
                assert key_valid_type_is_string is None
                assert key_valid_type_is_integer is None
            pytest.fail("We managed to get this far, which shouldn't be possible")
        except AssertionError:
            pass

        try:
            @vault.manual()
            def fn_faulty_default_assignment(key_valid_type_is_str=None, key_valid_type_is_int=None):
                assert key_valid_type_is_str == "valid"
                assert key_valid_type_is_int == 1
            pytest.fail("We managed to get this far, which shouldn't be possible")
        except AssertionError as e:
            pass

        try:
            @vault.manual()
            def fn_mixed_faulty_default_assignment(key_valid_type_is_str=None, key_valid_type_is_integer=varvault.AssignedByVault):
                assert key_valid_type_is_str == "valid"
                assert key_valid_type_is_integer == 1
            pytest.fail("We managed to get this far, which shouldn't be possible")
        except AssertionError as e:
            pass

    def test_create_vault_no_resource(self):
        vault = varvault.create(keyring=Keyring, name="vault")
        assert vault.resource is None
        vault.insert(Keyring.key_valid_type_is_str, "valid")
        vault.insert(Keyring.key_valid_type_is_int, 1)
        assert vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert vault.get(Keyring.key_valid_type_is_int) == 1

    def test_create_vault_with_json_resource_relative_path(self):
        try:
            vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(os.path.basename(vault_file_new), mode="w"))
            assert vault.resource
            assert vault.resource.path == os.path.basename(vault_file_new)
        finally:
            os.remove(os.path.basename(vault_file_new))

    def test_key_usages(self):
        class KeyringTemporary(varvault.Keyring):
            k1 = varvault.Key("k1")
            k2 = varvault.Key("k2")
            k3 = varvault.Key("k3")
        vault = varvault.create(keyring=KeyringTemporary, name="vault")
        @vault.manual(output=(KeyringTemporary.k1, KeyringTemporary.k2, KeyringTemporary.k3))
        def _set():
            return 1, "valid", 3.14

        try:
            vault.get(KeyringTemporary.k1)
            pytest.fail("We managed to get this far, which shouldn't be possible")
        except KeyError as e:
            logger.info(e.args[0])
            assert "test_vault._set" in e.args[0]

        _set()

        @vault.manual(input=(KeyringTemporary.k1, KeyringTemporary.k2))
        def _use_first(k1=varvault.AssignedByVault, k2=varvault.AssignedByVault):
            assert k1 == 1
            assert k2 == "valid"

        _use_first()

        @vault.manual(input=KeyringTemporary.k3)
        def _use_second(k3=varvault.AssignedByVault):
            assert k3 == 3.14

        _use_second()

        @vault.manual(varvault.Flags.permit_modifications, input=KeyringTemporary.k3, output=KeyringTemporary.k3)
        def _override(k3):
            assert k3 == 3.14
            return k3 * 2
        _override()

        assert vault.get(KeyringTemporary.k3) == 6.28
        assert KeyringTemporary.k1.usages.as_input == ['test_vault._use_first']
        assert KeyringTemporary.k2.usages.as_input == ['test_vault._use_first']
        assert KeyringTemporary.k3.usages.as_input == ['test_vault._override', 'test_vault._use_second']
        assert KeyringTemporary.k1.usages.as_return == ['test_vault._set']
        assert KeyringTemporary.k2.usages.as_return == ['test_vault._set']
        assert KeyringTemporary.k3.usages.as_return == ['test_vault._override', 'test_vault._set']

    def test_signature_with_missing_input_keys(self):
        vault = varvault.create(varvault.Flags.use_signature_for_input_keys,
                                keyring=Keyring,
                                name="vault",
                                resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(varvault.Flags.input_key_can_be_missing, input=Keyring.key_valid_type_is_str)
        def _use(key_valid_type_is_str=varvault.AssignedByVault):
            assert not key_valid_type_is_str
            assert key_valid_type_is_str is None, f"It's {type(key_valid_type_is_str)}, not {None}, which is what it should be"

        _use()

    def test_load_from_backup(self):
        assert not os.path.exists(vault_file_new)
        json.dump({Keyring.key_valid_type_is_str: "valid", Keyring.key_valid_type_is_int: 1}, open(f"{vault_file_new}.bak", "w"))
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="a"))

        assert vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert vault.get(Keyring.key_valid_type_is_int) == 1

        assert not os.path.exists(f"{vault_file_new}.bak"), "This file should have been renamed to remove extension '.bak'"

    def test_lambdas(self):
        vault = varvault.create(keyring=Keyring, name="vault")

        # Define the lambda function that will be used to set the value of the key
        setter: Callable = vault.lambdavaulter(lambda: 1, output_keys=Keyring.key_valid_type_is_int)
        assert callable(setter)

        # Define the lambda function that will be used to get the value of the key. Note that the name of the input variable is identical to the name of the key.
        getter: Callable = vault.lambdavaulter(lambda key_valid_type_is_int: key_valid_type_is_int * 3, input_keys=Keyring.key_valid_type_is_int)
        assert callable(getter)

        # Set the value of the key using the lambda
        setter()

        # Assert that the value of the key is as expected after setting with the lambda
        assert vault.get(Keyring.key_valid_type_is_int) == 1

        # Use the getter-lambda to get the value from the vault and assert that it is as expected
        assert getter() == 3

    def test_verify_faulty_extra_keys(self):
        try:
            vault = varvault.create(keyring=Keyring, name="vault", ex_1="ex_1", ex_2="ex_2")
            pytest.fail("We managed to get this far, which shouldn't be possible")
        except ValueError as e:
            assert "Faulty keys that must be changed: ['ex_1', 'ex_2']" in str(e.args[0]), e
        k1 = varvault.Key("ex1", valid_type=float)
        k2 = varvault.Key("ex2", valid_type=list)
        vault = varvault.create(keyring=Keyring, name="vault", k1=k1, k2=k2)
        vault.insert(k1, 3.14)
        vault.insert(k2, ["valid", "list", "of", "strings"])
        assert vault.get(k1) == 3.14
        assert vault.get(k2) == ["valid", "list", "of", "strings"]

    def test_modifiers(self):

        @varvault.modifier()
        def modifier_expand_path(value: str):
            return os.path.expanduser(value)

        @varvault.modifier()
        def modifier_multiply(value: int or float):
            return value * 2

        @varvault.modifier()
        def modifier_add(value: int or float):
            return value + 2

        @varvault.modifier()
        def modifier_failed(value: int or float):
            if value < 0:
                raise ValueError(f"Value {value} is less than 0")
            return value * 2
        try:
            @varvault.modifier()
            def modifier_no_return(value: int or float):
                value = value + 2
        except AssertionError as e:
            assert "No return-statement found in the functions source" in str(e.args[0]), e

        class KeyringModifiers(varvault.Keyring):
            key_path = varvault.Key("key_path", valid_type=str, modifiers=modifier_expand_path)
            key_int = varvault.Key("key_int", valid_type=int, modifiers=modifier_multiply)
            key_float = varvault.Key("key_float", valid_type=float, modifiers=(modifier_add, modifier_multiply))
            key_failed = varvault.Key("key_failed", valid_type=float, modifiers=modifier_failed)

        vault = varvault.create(keyring=KeyringModifiers, name="vault")
        vault.insert(KeyringModifiers.key_path, "~/.varvault")
        expected = os.path.expanduser("~/.varvault")
        assert vault.get(KeyringModifiers.key_path) == expected

        vault.insert(KeyringModifiers.key_int, 1)
        assert vault.get(KeyringModifiers.key_int) == 2, "The modifier should have doubled the value"

        vault.insert(KeyringModifiers.key_float, 1.0)
        assert vault.get(KeyringModifiers.key_float) == 6.0, "The modifiers should have added 2 and then doubled the value to 6.0"

        with pytest.raises(varvault.ModifierException) as e:
            vault.insert(KeyringModifiers.key_failed, -1.0)
        assert "Value -1.0 is less than 0" in str(e.value.args[0]), e

    def test_warning_write_to_read_only_resource(self):
        resource = varvault.JsonResource(vault_file_new, mode="w")
        resource.write({Keyring.key_valid_type_is_int: 1})

        vault = varvault.create(keyring=Keyring, name="vault", resource=varvault.JsonResource(vault_file_new, mode="r"))
        with warns(UserWarning, match="It appears you are trying to write to a resource that is not permitted to write to and the vault has already been initialized"):
            vault.insert(Keyring.key_valid_type_is_str, "should not be written")

        resource = varvault.JsonResource(vault_file_new, mode="r")
        assert Keyring.key_valid_type_is_str not in resource.create_mv(**Keyring.get_keys()), "The key should not have been written to the file"

        with warns(UserWarning, match="Tried to write to a resource defined as read-only"):
            resource.write({Keyring.key_valid_type_is_str: "should not be written"})
        assert Keyring.key_valid_type_is_str not in resource.create_mv(**Keyring.get_keys()), "The key should not have been written to the file"

    def test_faulty_put_on_vault(self):
        # For coverage
        vault = varvault.create(keyring=Keyring)
        with pytest.raises(NotImplementedError) as e:
            vault._put({Keyring.key_valid_type_is_str: "valid"})
        assert "Not implemented" in str(e.value.args[0]), e

    def test_vaulted_function_raises_exception(self):
        vault = varvault.create(keyring=Keyring)
        vault.insert(Keyring.key_valid_type_is_str, "valid")

        @vault.manual(input=Keyring.key_valid_type_is_str)
        async def async_func(id, key_valid_type_is_str: str = varvault.AssignedByVault):
            raise ValueError(key_valid_type_is_str)

        with pytest.raises(ValueError) as e:
            varvault.concurrent_execution(async_func, [1])
        assert "valid" in str(e.value.args[0]), e

        @vault.manual(input=Keyring.key_valid_type_is_str)
        def func(key_valid_type_is_str: str = varvault.AssignedByVault):
            raise ValueError(key_valid_type_is_str)

        with pytest.raises(ValueError) as e:
            func()
        assert "valid" in str(e.value.args[0]), e

    def test_no_error_logging_flag(self):
        vault = varvault.create(keyring=Keyring)
        vault.insert(Keyring.key_valid_type_is_str, "valid")

        @vault.manual(varvault.Flags.no_error_logging, input=Keyring.key_valid_type_is_str)
        async def async_func(id, key_valid_type_is_str: str = varvault.AssignedByVault):
            raise ValueError(key_valid_type_is_str)

        with pytest.raises(ValueError) as e:
            varvault.concurrent_execution(async_func, [1])
        assert "valid" in str(e.value.args[0]), e

        @vault.manual(varvault.Flags.no_error_logging, input=Keyring.key_valid_type_is_str)
        def func(key_valid_type_is_str: str = varvault.AssignedByVault):
            raise ValueError(key_valid_type_is_str)

        with pytest.raises(ValueError) as e:
            func()
        assert "valid" in str(e.value.args[0]), e

    def test_flag_is_invalid_type(self):
        vault = varvault.create(keyring=Keyring)
        with pytest.raises(TypeError) as e:
            vault.insert(Keyring.key_valid_type_is_str, "valid", varvault.Flags.permit_modifications.value)
        assert "is not of type" in str(e.value.args[0]), e

        with pytest.raises(TypeError) as e:
            @vault.manual(varvault.Flags.permit_modifications.value, output=Keyring.key_valid_type_is_str)
            def func():
                return "valid"
            func()

        assert "is not of type" in str(e.value.args[0]), e

    def test_return_values_cannot_be_none(self):
        class KeyringTemp:
            key_valid_type_is_str = varvault.Key("key_valid_type_is_str", valid_type=str, can_be_none=True)
        vault = varvault.create(keyring=Keyring)
        with pytest.raises(ValueError) as e:
            @vault.manual(varvault.Flags.return_values_cannot_be_none, output=KeyringTemp.key_valid_type_is_str)
            def func():
                return None
            func()
        assert "The value mapped to key_valid_type_is_str is None and Flags.return_values_cannot_be_none is defined" in str(e.value.args[0]), e

    def test_get(self):
        vault = varvault.create(keyring=Keyring)
        vault.insert(Keyring.key_valid_type_is_str, "valid")
        vault.insert(Keyring.key_valid_type_is_int, 1)

        key_valid_type_is_str = vault.get(Keyring.key_valid_type_is_str)
        assert key_valid_type_is_str == "valid", key_valid_type_is_str

        mv: varvault.MiniVault = vault.get([Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int])
        assert Keyring.key_valid_type_is_str in mv and Keyring.key_valid_type_is_int in mv, mv
        assert mv[Keyring.key_valid_type_is_str] == "valid" and mv[Keyring.key_valid_type_is_int] == 1, mv
        values = mv.values()
        assert "valid" in values and 1 in values, values

        with pytest.raises(NotImplementedError) as e:
            vault.get({Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int})
        assert re.match("Type .* is not supported for the 'get' method", str(e.value.args[0])) is not None, e

    def test_clean_no_valid_type(self):
        class Temp:
            def __init__(self, a1, a2):
                self.a1 = a1
                self.a2 = a2

        class KeyringTemp(varvault.Keyring):
            key_no_valid_type = varvault.Key("key_no_valid_type")
            key_special_valid_type = varvault.Key("key_special_valid_type", valid_type=Temp)

        vault = varvault.create(keyring=KeyringTemp)
        vault.insert(KeyringTemp.key_no_valid_type, "valid")
        vault.insert(KeyringTemp.key_special_valid_type, Temp("a1", "a2"))

        @vault.manual(varvault.Flags.clean_output_keys, output=(KeyringTemp.key_no_valid_type, KeyringTemp.key_special_valid_type))
        def clean():
            return

        clean()
        assert vault.get(KeyringTemp.key_no_valid_type) is None, vault.get(KeyringTemp.key_no_valid_type)
        assert vault.get(KeyringTemp.key_special_valid_type) is None, vault.get(KeyringTemp.key_special_valid_type)

    def test_validator_modifier_undecorated(self):
        def validate(value):
            return value == "valid"

        def modify(value):
            return value + " modified"

        with pytest.raises(ValueError) as e:
            key_validate = varvault.Key("key_validate", valid_type=str, validators=validate)
        assert "is not decorated with <function validator" in str(e.value.args[0]), e

        with pytest.raises(ValueError) as e:
            key_modify = varvault.Key("key_modify", valid_type=str, modifiers=modify)
        assert "is not decorated with <function modifier" in str(e.value.args[0]), e

    def test_compare_key_with_non_str(self):
        key = varvault.Key("key")
        assert not key == 1, key

    def test_valid_type_corner_cases(self):
        class Temp:
            pass

        class KeyringTemp(varvault.Keyring):
            key_is_type_can_be_none = varvault.Key("key_is_type_can_be_none", valid_type=Temp, can_be_none=True)
            key_is_type_cannot_be_none = varvault.Key("key_is_type_cannot_be_none", valid_type=Temp)
            key_valid_type_is_int = varvault.Key("key_valid_type_is_int", valid_type=int)

        vault = varvault.create(varvault.Flags.permit_modifications, keyring=KeyringTemp)
        vault.insert(KeyringTemp.key_is_type_can_be_none, Temp)

        with pytest.raises(ValueError) as e:
            vault.insert(KeyringTemp.key_is_type_can_be_none, dict())
        assert "requires type to be" in str(e.value.args[0]), e
        vault.insert(KeyringTemp.key_is_type_cannot_be_none, Temp)
        assert vault.get(KeyringTemp.key_is_type_cannot_be_none) is Temp, vault.get(KeyringTemp.key_is_type_cannot_be_none)
        assert vault.get(KeyringTemp.key_is_type_can_be_none) is Temp, vault.get(KeyringTemp.key_is_type_can_be_none)

    def test_get_non_matching_key(self):
        class KeyringTemp(varvault.Keyring):
            key_valid_type_is_int = varvault.Key("key_valid_type_is_int", valid_type=int)
        with pytest.raises(KeyError) as e:
            KeyringTemp.get_key_by_matching_string("not_valid")
        assert "Failed to get matching key for string: not_valid" in str(e.value.args[0]), e

    def test_validator_bool_annotate_warning(self):
        with warns(SyntaxWarning) as w:
            @varvault.validator(function_returns_bool=True)
            def validate(value):  # Note lack of annotated bool return type
                return value == "valid"
        assert "If you define the validator-function to say that it returns a bool, you should really also annotate the function to say it returns a bool." in str(w[0].message), w

    def test_try_modify_mode_properties(self):
        resource = varvault.JsonResource(existing_vault, mode=varvault.ResourceModes.READ)
        with pytest.raises(AttributeError) as e:
            resource.mode_properties.load = False  # Piss off, you can't modify mode_properties after they've been created
        assert "ModeProperties are immutable" in str(e.value.args[0]), e

    def test_resource_invalid_mode(self):
        with pytest.raises(ValueError) as e:
            resource = varvault.JsonResource(existing_vault, mode="invalid")
        assert "Invalid mode: invalid" in str(e.value.args[0]), e

    def test_non_existing_resource(self):
        with pytest.raises(varvault.ResourceNotFoundError) as e:
            resource = varvault.JsonResource("non_existing_resource.json")
        assert "Resource not found at:" in str(e.value.args[0]), e

    def test_resource_invalid_value_write(self):
        class Temp:
            pass

        resource = varvault.JsonResource(vault_file_new, mode="w")
        with pytest.raises(varvault.ResourceNotFoundError) as e:
            d = dict(k=Temp)
            assert not resource.writable(d)
            resource.write(d)
        assert "Failed to write to the resource:" in str(e.value.args[0]), e

    def test_resource_invalid_value_read(self):
        resource = varvault.JsonResource(vault_file_new, mode="w")
        resource.create()
        open(vault_file_new, "w").write("invalid")

        try:
            resource.read()
        except varvault.ResourceNotFoundError as e:
            assert "Failed to read from the resource" in str(e), e
        resource = varvault.JsonResource(vault_file_new, mode="r")
        with pytest.raises(varvault.ResourceNotFoundError) as e:
            resource.create()
        assert "Unable to read from resource at" in str(e.value.args[0]), e

    def test_append_resource_mode(self):
        resource = varvault.JsonResource(vault_file_new, mode="a")
        resource.create()
        resource.write({"key": "value"})
        resource = varvault.JsonResource(vault_file_new, mode="r")
        assert resource.read() == {"key": "value"}, resource.read()

    def test_modifier_plus_validator_on_insert(self):
        @varvault.validator(function_returns_bool=True)
        def no_dashes_in_str(value: str) -> bool:
            return re.match(r"^[a-zA-Z0-9_]*$", value) is not None

        @varvault.modifier()
        def remove_dashes_in_str(value: str) -> str:
            return value.replace("-", "_")

        @varvault.modifier()
        def remove_underscore_in_str(value: str) -> str:
            return value.replace("_", "-")

        @varvault.validator(function_returns_bool=True)
        def no_underscore_in_str(value: str) -> bool:
            return re.match(r"^[a-zA-Z0-9-]*$", value) is not None

        class KeyringKeyValidationFunction(varvault.Keyring):
            name = varvault.Key("name", valid_type=str, modifiers=remove_underscore_in_str, validators=no_underscore_in_str)
            path = varvault.Key("path", valid_type=str, modifiers=remove_dashes_in_str, validators=no_dashes_in_str)

        vault = varvault.create(varvault.Flags.permit_modifications, keyring=KeyringKeyValidationFunction)
        vault.insert(KeyringKeyValidationFunction.name, "will_be_modified_to_dashes")
        vault.insert(KeyringKeyValidationFunction.path, "will-be-modified-to-underscores")
        assert vault.get(KeyringKeyValidationFunction.name) == "will-be-modified-to-dashes", vault.get(KeyringKeyValidationFunction.name)
        assert vault.get(KeyringKeyValidationFunction.path) == "will_be_modified_to_underscores", vault.get(KeyringKeyValidationFunction.path)




import json
import tempfile
import threading

from commons import *


vault_file_new = f"{DIR}/new-vault.json"
vault_file_new_secondary = f"{DIR}/new-vault-secondary.json"


class TestLiveUpdate:

    @classmethod
    def setup_class(cls):
        tempfile.tempdir = "/tmp" if sys.platform == "darwin" or sys.platform == "linux" else tempfile.gettempdir()

    def setup_method(self):
        try:
            os.remove(vault_file_new)
        except:
            pass
        try:
            os.remove(vault_file_new_secondary)
        except:
            pass
        try:
            os.remove(vault_file_new + ".bak")
        except:
            pass

    def test_live_update_vault(self):
        vault_new = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w+"))
        vault_from = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="r+"))

        @vault_new.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()

        assert Keyring.key_valid_type_is_str not in vault_from, f"{Keyring.key_valid_type_is_str} already in the vault; This should not be the case"

        @vault_from.manual(input=Keyring.key_valid_type_is_str)
        def _get(**kwargs):
            v = kwargs.get(Keyring.key_valid_type_is_str)
            assert v == "valid", f"Value {v} is not correct; Live-update doesn't work"
        _get()

    def test_live_update_on_main_vault(self):
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w+"))

        @vault.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()

        assert Keyring.key_valid_type_is_int not in vault, f"{Keyring.key_valid_type_is_int} already in vault; This should not be possible"

        vault_data = json.load(open(vault_file_new))
        vault_data[Keyring.key_valid_type_is_int] = 1
        json.dump(vault_data, open(vault_file_new, "w"), indent=2)

        assert Keyring.key_valid_type_is_int not in vault, f"{Keyring.key_valid_type_is_int} already in vault; This should not be possible"

        @vault.manual(input=Keyring.key_valid_type_is_int)
        def _get(**kwargs):
            key_valid_type_is_int = kwargs.get(Keyring.key_valid_type_is_int)
            assert key_valid_type_is_int == 1
        _get()

    def test_create_live_update_vault(self):
        assert not os.path.exists(vault_file_new)
        # The file should NOT be created here
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="r+"))
        assert not os.path.exists(vault_file_new)
        # Here the file SHOULD be created
        vault = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w+"))
        assert os.path.exists(vault_file_new)
        assert Keyring.key_valid_type_is_str not in vault
        assert Keyring.key_valid_type_is_int not in vault

        json.dump({Keyring.key_valid_type_is_str: "valid", Keyring.key_valid_type_is_int: 1}, open(vault_file_new, "w"))

        assert vault.get(Keyring.key_valid_type_is_str) == "valid"
        assert vault.get(Keyring.key_valid_type_is_int) == 1

    def test_threaded_vaults_live_update(self):
        def runner():
            vault_thread = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
            vault_thread.insert(Keyring.key_valid_type_is_str, "valid")
            vault_thread.insert(Keyring.key_valid_type_is_int, 1)

        assert not os.path.exists(vault_file_new)
        vault_outer = varvault.create(keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="r+"))
        assert not os.path.exists(vault_file_new), "Vault file is created when it shouldn't be"

        t = threading.Thread(target=runner)
        t.start()
        t.join()

        @vault_outer.manual(input=(Keyring.key_valid_type_is_str, Keyring.key_valid_type_is_int))
        def validate(key_valid_type_is_str=None, key_valid_type_is_int=None):
            assert key_valid_type_is_str == "valid"
            assert key_valid_type_is_int == 1
        validate()

    def test_resource_live_update(self):
        resource = varvault.JsonResource(vault_file_new, mode="w+")
        assert not resource.exists()
        resource.create()
        assert resource.exists()
        pre_state = resource.state
        json.dump({Keyring.key_valid_type_is_str: "valid", Keyring.key_valid_type_is_int: 1}, open(vault_file_new, "w"))
        assert resource.resource_has_changed()
        assert resource.cached_state != pre_state
        data = resource.read()
        assert data[Keyring.key_valid_type_is_str] == "valid"
        assert data[Keyring.key_valid_type_is_int] == 1
        resource.update_state()

        assert resource.last_known_state == resource.cached_state



from typing import *

from .utils import assert_and_raise
from .keyring import Key


class MiniVault(dict):
    """Class that acts as a very bare-bones vault to represent variables either coming from the vault via get_multiple, or when returning variables form a vaulted function."""

    def __init__(self, d=None, **kwargs):
        d = d or {}
        super(MiniVault, self).__init__(d, **kwargs)

    @staticmethod
    def build(keys: List[Key] or Tuple[Key], values: List[Any] or Tuple[Any]):
        f"""Builds a {MiniVault}-object based on iterables of {keys} and {values}. Note that the number of {keys} and the number of {values} must be identical"""
        assert isinstance(keys, (list, tuple)), f"'keys' must be a list or a tuple, not {type(keys)}"
        assert isinstance(values, (list, tuple)), f"'values' must be a list or a tuple, not {type(values)}"

        assert len(keys) == len(values), "The length of 'keys' and 'values' must be identical"

        data = dict()
        for i in range(len(keys)):
            key = keys[i]
            value = values[i]
            assert isinstance(key, Key), f"Key {key} is not of correct type {Key}"
            data[key] = value

        return MiniVault(data)

    def add(self, k, v):
        """Adds a value mapped to a key. Essentially just wraps the '__setitem__' function."""
        self.__setitem__(k, v)

    def keys(self) -> List[Key]:
        """Returns a list of keys"""
        return list(super(MiniVault, self).keys())

    def values(self) -> List[Any]:
        """Returns a list of values"""
        return list(super(MiniVault, self).values())



import asyncio
import enum
from types import *
from typing import *


class AssignedByVaultEnum(enum.Enum):
    """
    Enum that represents the assigned_by_vault attribute of a variable.
    """
    ASSIGNED = True


AssignedByVault = AssignedByVaultEnum.ASSIGNED


def concurrent_execution(target: Union[Coroutine, FunctionType, Callable], *inputs, **kwargs):
    """
    Wraps the asyncio API in Python to make it a bit easier to work with.

    Example usage:
    ```
    async def run(arg1, arg2, const=None):
        assert const == "2"
        print(arg1, arg2)
        return (arg1 + arg2) * int(const)

    values = concurrent_execution(run, [1, 2, 3, 4, 5], [10, 20, 30, 40, 50], const="2")
    print(values)
    ```
    Output will be
    ```
    [22, 44, 66, 88, 110]
    ```

    :param target: A callable defined as coroutine via the keyword "async" that takes an arbitrary amount of arguments
    :param inputs: An arbitrary tuple of arguments as iterables. All iterables will be zipped together like this: zipped = list(zip(*inputs))
    :param kwargs: Kwargs that are treated like constants that will be sent to each call of 'target'. Any object in kwargs will NOT be zipped into the other arguments.
    :return: Whatever target returns, but as a list of what it returned.
    """
    assert asyncio.iscoroutinefunction(target), f"'target' ({target.__module__}.{target.__name__}) is not a coroutine function; define it as 'async def {target.__name__}(...)'"

    async def do(_target, *_inputs, **_kwargs):
        zipped = list(zip(*_inputs))
        _ret = await asyncio.gather(*[asyncio.create_task(_target(*i, **_kwargs)) for i in zipped])
        return _ret

    return asyncio.run(do(target, *inputs, **kwargs))


def assert_and_raise(condition: bool, exception: BaseException):
    """
    Asserts a condition and raises an exception if the condition is False.
    Note: Checks if __debug__ is set to True before running the assertion, just how builtin `assert` works.

    :param condition: The condition to assert
    :param exception: The exception to raise
    """
    if __debug__ and not condition:
        raise exception



import json
import time

from commons import *

vault_file_new = f"{DIR}/new-vault.json"
large_vault_file = f"{DIR}/large-scale-existing-vault.json"


class ResultToUploadDict(varvault.VaultStructDictBase):
    def __init__(self, **vault_pairs):
        super(ResultToUploadDict, self).__init__(**vault_pairs)
        self.nsover = vault_pairs.get("nsover")
        self.buildid = vault_pairs.get("buildid")
        self.testid = vault_pairs.get("testid")
        self.testdescr = vault_pairs.get("testdescr")
        self.starttime = vault_pairs.get("starttime")
        self.endtime = vault_pairs.get("endtime")
        self.status = vault_pairs.get("status")
        self.logpath = vault_pairs.get("logpath")

    @classmethod
    def create(cls, vault_key, vault_value):
        return ResultToUploadDict(**vault_value)


class ResultDict(ResultToUploadDict):
    def __init__(self, **vault_pairs):
        super(ResultDict, self).__init__(**vault_pairs)
        self.version_ha1 = vault_pairs.get("version_ha1")
        self.version_ha2 = vault_pairs.get("version_ha2")
        self.version_ha3 = vault_pairs.get("version_ha3")

    @classmethod
    def create(cls, vault_key, vault_value):
        return ResultDict(**vault_value)


class GroupInstallers(varvault.VaultStructDictBase):
    class Group(varvault.VaultStructDictBase):
        def __init__(self, installer, installer_upgrade=None):
            super(GroupInstallers.Group, self).__init__(installer=installer, installer_upgrade=installer_upgrade)
            self.installer = installer
            self.installer_upgrade = installer_upgrade

        @classmethod
        def create(cls, vault_key, vault_value):
            return GroupInstallers.Group(**vault_value)

    def __init__(self, **vault_pairs):
        super(GroupInstallers, self).__init__(**vault_pairs)
        for group, group_data in vault_pairs.items():
            self.__setattr__(group, self.Group.create(vault_key=group,
                                                      vault_value=dict(installer=group_data.get("installer"),
                                                                       installer_upgrade=group_data.get("installer-upgrade"))))

    @classmethod
    def create(cls, vault_key, vault_value):
        return GroupInstallers(**vault_value)


class KeyringLargeScale(varvault.Keyring):
    logname = varvault.Key("logname", valid_type=str)
    log_out_dir = varvault.Key("log_out_dir", valid_type=str)
    hostsfile_path = varvault.Key("hostsfile_path", valid_type=str)
    influxdb_ip = varvault.Key("influxdb_ip", valid_type=str)
    influxdb_port = varvault.Key("influxdb_port", valid_type=str)
    influxdb_database = varvault.Key("influxdb_database", valid_type=str)
    grafana_hostname = varvault.Key("grafana_hostname", valid_type=str)
    grafana_port = varvault.Key("grafana_port", valid_type=str)
    grafana_credentials = varvault.Key("grafana_credentials", valid_type=str)
    grafana_auth_token = varvault.Key("grafana_auth_token", valid_type=str)
    result = varvault.Key("result", valid_type=ResultDict)
    result_to_upload = varvault.Key("result_to_upload", valid_type=ResultToUploadDict)
    installer_local = varvault.Key("installer_local", valid_type=str)
    installer_local_pre_upgrade = varvault.Key("installer_local_pre_upgrade", valid_type=str)
    installer_local_post_upgrade = varvault.Key("installer_local_post_upgrade", valid_type=str)
    group_installers = varvault.Key("group_installers", valid_type=GroupInstallers)
    descriptor_file = varvault.Key("descriptor_file", valid_type=str)
    docker_network = varvault.Key("docker_network", valid_type=str)
    docker_network_existed_on_start = varvault.Key("docker_network_existed_on_start", valid_type=bool)
    cpus_per_container = varvault.Key("cpus_per_container", valid_type=int, can_be_none=True)
    target_env = varvault.Key("target_env", valid_type=str)
    workspace_dir = varvault.Key("workspace_dir", valid_type=str)
    test_time = varvault.Key("test_time", valid_type=str)
    start_time = varvault.Key("start_time", valid_type=float)
    end_time = varvault.Key("end_time", valid_type=float)
    build_id = varvault.Key("build_id", valid_type=str)
    test_id = varvault.Key("test_id", valid_type=str)
    test_description = varvault.Key("test_description", valid_type=str)
    instance_count = varvault.Key("instance_count", valid_type=str)
    container_count = varvault.Key("container_count", valid_type=str)
    version = varvault.Key("version", valid_type=str)
    branch = varvault.Key("branch", valid_type=str)
    branch_custom = varvault.Key("branch_custom", valid_type=str, can_be_none=True)
    branch_ha1 = varvault.Key("branch_ha1", valid_type=str)
    branch_ha2 = varvault.Key("branch_ha2", valid_type=str)
    branch_ha3 = varvault.Key("branch_ha3", valid_type=str)
    version_ha1 = varvault.Key("version_ha1", valid_type=str)
    version_ha2 = varvault.Key("version_ha2", valid_type=str)
    version_ha3 = varvault.Key("version_ha3", valid_type=str)
    branch_upgrade_ha1 = varvault.Key("branch_upgrade_ha1", valid_type=str)
    branch_upgrade_ha2 = varvault.Key("branch_upgrade_ha2", valid_type=str)
    branch_upgrade_ha3 = varvault.Key("branch_upgrade_ha3", valid_type=str)
    version_upgrade_ha1 = varvault.Key("version_upgrade_ha1", valid_type=str)
    version_upgrade_ha2 = varvault.Key("version_upgrade_ha2", valid_type=str)
    version_upgrade_ha3 = varvault.Key("version_upgrade_ha3", valid_type=str)
    version_pre_upgrade = varvault.Key("version_pre_upgrade", valid_type=str)
    validation_errors = varvault.Key("validation_errors", valid_type=list, can_be_none=True)
    validation_result = varvault.Key("validation_result", valid_type=dict)


class TestLargeScaleVault:

    def setup_method(self):
        try:
            os.remove(vault_file_new)
        except:
            pass

    def test_load_from_large_vault(self):
        resource_mv = varvault.JsonResource(large_vault_file, mode="r").create_mv(**KeyringLargeScale.get_keys())
        vault = varvault.create(varvault.Flags.remove_existing_log_file, keyring=KeyringLargeScale, resource=varvault.JsonResource(vault_file_new, "w"))
        vault.insert_minivault(resource_mv)
        logger.info("Created the vault from file")
        # We load from existing and write to new, so just check that the new file contains the correct data
        contents = json.load(open(vault_file_new))
        assert KeyringLargeScale.workspace_dir in contents
        assert KeyringLargeScale.test_time in contents
        assert KeyringLargeScale.cpus_per_container in contents
        assert KeyringLargeScale.docker_network in contents
        assert KeyringLargeScale.branch in contents
        assert KeyringLargeScale.descriptor_file in contents
        assert KeyringLargeScale.logname in contents
        assert KeyringLargeScale.log_out_dir in contents
        assert KeyringLargeScale.hostsfile_path in contents
        assert KeyringLargeScale.grafana_port in contents
        assert KeyringLargeScale.influxdb_database in contents
        assert KeyringLargeScale.installer_local_pre_upgrade in contents
        assert KeyringLargeScale.installer_local_post_upgrade in contents
        assert KeyringLargeScale.group_installers in contents
        assert KeyringLargeScale.branch_custom in contents
        assert KeyringLargeScale.target_env in contents
        assert KeyringLargeScale.start_time in contents
        assert KeyringLargeScale.version in contents
        assert KeyringLargeScale.version_pre_upgrade in contents
        assert KeyringLargeScale.version_ha1 in contents
        assert KeyringLargeScale.version_ha2 in contents
        assert KeyringLargeScale.version_ha3 in contents
        assert KeyringLargeScale.version_upgrade_ha1 in contents
        assert KeyringLargeScale.version_upgrade_ha2 in contents
        assert KeyringLargeScale.version_upgrade_ha3 in contents
        assert KeyringLargeScale.build_id in contents
        assert KeyringLargeScale.test_id in contents
        assert KeyringLargeScale.test_description in contents
        assert KeyringLargeScale.branch_ha1 in contents
        assert KeyringLargeScale.branch_ha2 in contents
        assert KeyringLargeScale.branch_ha3 in contents
        assert KeyringLargeScale.branch_upgrade_ha1 in contents
        assert KeyringLargeScale.branch_upgrade_ha2 in contents
        assert KeyringLargeScale.branch_upgrade_ha3 in contents
        assert KeyringLargeScale.docker_network_existed_on_start in contents
        assert KeyringLargeScale.influxdb_ip in contents
        assert KeyringLargeScale.grafana_hostname in contents
        assert KeyringLargeScale.end_time in contents
        assert KeyringLargeScale.validation_result in contents
        assert KeyringLargeScale.validation_errors in contents
        assert KeyringLargeScale.result_to_upload in contents
        assert KeyringLargeScale.result in contents

    def test_get_multiple(self):
        resource_mv = varvault.JsonResource(large_vault_file, mode="r").create_mv(**KeyringLargeScale.get_keys())
        vault = varvault.create(varvault.Flags.remove_existing_log_file, keyring=KeyringLargeScale, resource=varvault.JsonResource(vault_file_new, "w"))
        vault.insert_minivault(resource_mv)
        keys = [KeyringLargeScale.workspace_dir,
                KeyringLargeScale.test_time,
                KeyringLargeScale.start_time,
                KeyringLargeScale.grafana_port,
                KeyringLargeScale.group_installers,
                KeyringLargeScale.version_pre_upgrade,
                KeyringLargeScale.version_upgrade_ha1,
                KeyringLargeScale.version_upgrade_ha2,
                KeyringLargeScale.version_upgrade_ha3,
                KeyringLargeScale.result_to_upload,
                KeyringLargeScale.result,
                ]

        vars = vault.get(keys)
        # This list should be empty; If not, a key is missing in the vault
        assert len([key for key in keys if key not in vars]) == 0, f"Keys are missing in vars: {[key for key in keys if key not in vars]}"

    def test_get_via_vaulter(self):
        resource_mv = varvault.JsonResource(large_vault_file, mode="r").create_mv(**KeyringLargeScale.get_keys())
        vault = varvault.create(varvault.Flags.remove_existing_log_file, keyring=KeyringLargeScale, resource=varvault.JsonResource(vault_file_new, "w"))
        vault.insert_minivault(resource_mv)
        keys = [KeyringLargeScale.workspace_dir,
                KeyringLargeScale.test_time,
                KeyringLargeScale.cpus_per_container,
                KeyringLargeScale.docker_network,
                KeyringLargeScale.branch,
                KeyringLargeScale.descriptor_file,
                KeyringLargeScale.logname,
                KeyringLargeScale.log_out_dir,
                KeyringLargeScale.hostsfile_path,
                KeyringLargeScale.grafana_port,
                KeyringLargeScale.influxdb_database,
                KeyringLargeScale.installer_local_pre_upgrade,
                KeyringLargeScale.installer_local_post_upgrade,
                KeyringLargeScale.group_installers,
                KeyringLargeScale.branch_custom,
                KeyringLargeScale.target_env,
                KeyringLargeScale.start_time,
                KeyringLargeScale.version,
                KeyringLargeScale.version_pre_upgrade,
                KeyringLargeScale.version_ha1,
                KeyringLargeScale.version_ha2,
                KeyringLargeScale.version_ha3,
                KeyringLargeScale.version_upgrade_ha1,
                KeyringLargeScale.version_upgrade_ha2,
                KeyringLargeScale.version_upgrade_ha3,
                KeyringLargeScale.build_id,
                KeyringLargeScale.test_id,
                KeyringLargeScale.test_description,
                KeyringLargeScale.branch_ha1,
                KeyringLargeScale.branch_ha2,
                KeyringLargeScale.branch_ha3,
                KeyringLargeScale.branch_upgrade_ha1,
                KeyringLargeScale.branch_upgrade_ha2,
                KeyringLargeScale.branch_upgrade_ha3,
                KeyringLargeScale.docker_network_existed_on_start,
                KeyringLargeScale.influxdb_ip,
                KeyringLargeScale.grafana_hostname,
                KeyringLargeScale.end_time,
                KeyringLargeScale.validation_result,
                KeyringLargeScale.validation_errors,
                KeyringLargeScale.result_to_upload,
                KeyringLargeScale.result,
                ]

        @vault.manual(input=keys)
        def _get(**kwargs):
            assert len([k for k in keys if k in kwargs]) == len(keys)
        _get()

    def test_get_threaded(self):
        resource_mv = varvault.JsonResource(large_vault_file, mode="r").create_mv(**KeyringLargeScale.get_keys())
        vault = varvault.create(varvault.Flags.remove_existing_log_file, keyring=KeyringLargeScale, resource=varvault.JsonResource(vault_file_new, "w"))
        vault.insert_minivault(resource_mv)
        keys = [KeyringLargeScale.workspace_dir,
                KeyringLargeScale.test_time,
                KeyringLargeScale.cpus_per_container,
                KeyringLargeScale.docker_network,
                KeyringLargeScale.branch,
                KeyringLargeScale.descriptor_file,
                KeyringLargeScale.logname,
                KeyringLargeScale.log_out_dir,
                KeyringLargeScale.hostsfile_path,
                KeyringLargeScale.grafana_port,
                KeyringLargeScale.influxdb_database,
                KeyringLargeScale.installer_local_pre_upgrade,
                KeyringLargeScale.installer_local_post_upgrade,
                KeyringLargeScale.group_installers,
                KeyringLargeScale.branch_custom,
                KeyringLargeScale.target_env,
                KeyringLargeScale.start_time,
                KeyringLargeScale.version,
                KeyringLargeScale.version_pre_upgrade,
                KeyringLargeScale.version_ha1,
                KeyringLargeScale.version_ha2,
                KeyringLargeScale.version_ha3,
                KeyringLargeScale.version_upgrade_ha1,
                KeyringLargeScale.version_upgrade_ha2,
                KeyringLargeScale.version_upgrade_ha3,
                KeyringLargeScale.build_id,
                KeyringLargeScale.test_id,
                KeyringLargeScale.test_description,
                KeyringLargeScale.branch_ha1,
                KeyringLargeScale.branch_ha2,
                KeyringLargeScale.branch_ha3,
                KeyringLargeScale.branch_upgrade_ha1,
                KeyringLargeScale.branch_upgrade_ha2,
                KeyringLargeScale.branch_upgrade_ha3,
                KeyringLargeScale.docker_network_existed_on_start,
                KeyringLargeScale.influxdb_ip,
                KeyringLargeScale.grafana_hostname,
                KeyringLargeScale.end_time,
                KeyringLargeScale.validation_result,
                KeyringLargeScale.validation_errors,
                KeyringLargeScale.result_to_upload,
                KeyringLargeScale.result,
                ]

        # Note the use of varvault.Flags.silent here. It will speed up the processing significantly.
        # It brings processing down from 3 seconds to 0.6 seconds for 500 parallel requests.
        @vault.manual(varvault.Flags.silent, input=keys)
        async def _get(thread_id, **kwargs):
            assert isinstance(thread_id, int)
            assert len([k for k in keys if k in kwargs]) == len(keys)

        num_threads = 500
        thread_ids = [i for i in range(num_threads)]
        logger.info(f"Number of threads: {len(thread_ids)}")
        start = time.time()
        varvault.concurrent_execution(_get, thread_ids)
        logger.info(f"Time (seconds): {time.time() - start}")

    def test_live_update(self):
        resource_mv = varvault.JsonResource(large_vault_file, mode="r").create_mv(**KeyringLargeScale.get_keys())
        vault = varvault.create(varvault.Flags.remove_existing_log_file, keyring=KeyringLargeScale, resource=varvault.JsonResource(vault_file_new, "w+"))
        vault.insert_minivault(resource_mv)

        keys = [KeyringLargeScale.workspace_dir,
                KeyringLargeScale.test_time,
                KeyringLargeScale.cpus_per_container,
                KeyringLargeScale.docker_network,
                KeyringLargeScale.branch,
                KeyringLargeScale.descriptor_file,
                KeyringLargeScale.logname,
                KeyringLargeScale.log_out_dir,
                KeyringLargeScale.hostsfile_path,
                KeyringLargeScale.grafana_port,
                KeyringLargeScale.influxdb_database,
                KeyringLargeScale.installer_local_pre_upgrade,
                KeyringLargeScale.installer_local_post_upgrade,
                KeyringLargeScale.group_installers,
                KeyringLargeScale.branch_custom,
                KeyringLargeScale.target_env,
                KeyringLargeScale.start_time,
                KeyringLargeScale.version,
                KeyringLargeScale.version_pre_upgrade,
                KeyringLargeScale.version_ha1,
                KeyringLargeScale.version_ha2,
                KeyringLargeScale.version_ha3,
                KeyringLargeScale.version_upgrade_ha1,
                KeyringLargeScale.version_upgrade_ha2,
                KeyringLargeScale.version_upgrade_ha3,
                KeyringLargeScale.build_id,
                KeyringLargeScale.test_id,
                KeyringLargeScale.test_description,
                KeyringLargeScale.branch_ha1,
                KeyringLargeScale.branch_ha2,
                KeyringLargeScale.branch_ha3,
                KeyringLargeScale.branch_upgrade_ha1,
                KeyringLargeScale.branch_upgrade_ha2,
                KeyringLargeScale.branch_upgrade_ha3,
                KeyringLargeScale.docker_network_existed_on_start,
                KeyringLargeScale.influxdb_ip,
                KeyringLargeScale.grafana_hostname,
                KeyringLargeScale.end_time,
                KeyringLargeScale.validation_result,
                KeyringLargeScale.validation_errors,
                KeyringLargeScale.result_to_upload,
                KeyringLargeScale.result,
                ]

        # Load the existing vault file and write the contents to the temporary file
        contents = json.load(open(large_vault_file))
        json.dump(contents, open(vault_file_new, "w"), indent=2)

        # Since we use live update, the contents of the existing vault file should now be loaded into the vault.
        @vault.manual(input=keys)
        def _get(**kwargs):
            assert len([k for k in keys if k in kwargs]) == len(keys)
        _get()



import logging
import warnings

from typing import Type

from .resource import BaseResource
from .keyring import Keyring, Key
from .flags import Flags
from .vault import VarVault


def create(*flags: Flags,
           keyring: Type[Keyring] = None,
           name: str = None,
           resource: BaseResource = None,
           logger: logging.Logger = None,
           **extra_keys: Key) -> VarVault:
    f"""
    Factory-function to help create a Vault-object instead of creating it manually (which is still possible).

    :param keyring: Describes which keys belong to this Vault.
    :param name: Optional. Used to name the vault and the logger for writing information and debug messages, unless {logger} was passed.
    :param resource: Optional resource to read/write the vault from/to respectively.
    :param flags: Optional arguments for defining {Flags} for this Vault.
     Note that any global {Flags} will be overridden by {Flags} defined in a vault-decorator.
    :param logger: Optional argument for defining your own logger object if you want to use a specific logger rather than varvault's own logger.
    :param extra_keys: Extra keys as a dict to write to. These keys can be defined during runtime, which can sometimes be necessary.
     It is recommended to use pre-determined keys (e.g. constants), but sometimes being more flexible can be useful.
    :return: Vault object based on input to this function.
    """
    faulty_keys = [k for k, v in extra_keys.items() if not isinstance(v, Key)]
    if faulty_keys:
        raise ValueError(f"Extra keys must be of type Key. Faulty keys that must be changed: {faulty_keys}")
    extra_keys_cleaned = {str(v.key_name): v for _, v in extra_keys.items()}
    extra_keys = extra_keys_cleaned
    initial_vars = None
    if resource and resource.mode_properties.load:
        vault_file_data = resource.read()

        assert isinstance(vault_file_data, dict), f"It appears we were not able to load a Dict from {resource}. Are you sure this is a valid resource? (content={vault_file_data})"
        keys_in_keyring = keyring.get_keys()
        keys_in_keyring.update(extra_keys)
        keys_in_file = vault_file_data.keys()
        _keys_not_in_keyring = [k for k in keys_in_file if k not in keys_in_keyring]
        if _keys_not_in_keyring and not Flags.is_set((Flags.ignore_keys_not_in_keyring,), *flags) and not resource.mode_properties.read_only:
            raise ValueError(f"Some keys in the resource were not in the keyring: {_keys_not_in_keyring}\n"
                             f"You can set to ignore this through the flag {Flags.ignore_keys_not_in_keyring}, "
                             f"or by setting the resource's mode property to one that includes 'read_only'")
        elif _keys_not_in_keyring and Flags.is_set((Flags.ignore_keys_not_in_keyring,), *flags):
            warnings.warn(f"Keys were found in the resource '{resource}' that are not in the keyring, "
                          f"but you have set to ignore keys found that are not in the keyring. "
                          f"These ignored keys will not be loaded to the vault.\n"
                          f"Keys not in the keyring: {_keys_not_in_keyring}")
        initial_vars = resource.create_mv(**keys_in_keyring)

    vault = VarVault(*flags,
                     keyring=keyring,
                     name=name,
                     resource=resource,
                     logger=logger,
                     initial_vars=initial_vars,
                     **extra_keys)

    return vault



import os
import logging
import tempfile


def get_logger(name, remove_existing_log_file=False):
    f"""Returns a logger for {name} that logs to a specific file."""

    class CustomFormat(logging.Formatter):
        FORMATS = {logging.INFO: logging.PercentStyle(f'%(levelname)s - %(asctime)s %(name)s - %(message)s'),
                   logging.DEBUG: logging.PercentStyle(f'%(levelname)s - %(asctime)s %(name)s - %(message)s'),
                   logging.WARNING: logging.PercentStyle(f'%(levelname)s - %(asctime)s %(name)s - %(message)s'),
                   logging.ERROR: logging.PercentStyle(f'%(levelname)s - %(asctime)s %(name)s - %(message)s'),
                   "DEFAULT": logging.PercentStyle(f'%(levelname)s - %(asctime)s %(name)s - %(message)s')}

        def __init__(self):
            default_style = CustomFormat.FORMATS["DEFAULT"]
            super(CustomFormat, self).__init__(fmt=default_style._fmt, style="%")

        def format(self, record):

            self._style = CustomFormat.FORMATS.get(record.levelno, CustomFormat.FORMATS["DEFAULT"])
            result = logging.Formatter.format(self, record)
            return result

    name = f"varvault-{name}" if name else "varvault"
    temp_file = tempfile.NamedTemporaryFile(suffix=".log")
    temp_file.close()
    temp_dir = os.path.dirname(temp_file.name)
    temp_file.name = os.path.join(temp_dir, "varvault-logs", f"{name}.log")
    os.makedirs(os.path.dirname(temp_file.name), exist_ok=True)

    log_file = temp_file.name
    if remove_existing_log_file:
        try:
            os.remove(log_file)
        except:
            pass

    log = logging.getLogger(name)
    log.handlers.clear()

    formatter = CustomFormat()

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    log.addHandler(ch)

    ch = logging.FileHandler(filename=log_file)
    ch.setFormatter(formatter)
    log.addHandler(ch)

    configure_logger(log)

    log.debug(f"Logger '{log.name}' created and logging to file '{log_file}'.")

    return log


def configure_logger(logger: logging.Logger, overall_level=logging.DEBUG, stream_level=logging.INFO, file_level=logging.DEBUG):
    f"""Configures a {logger} by setting the logging levels for the different handlers."""
    logger.setLevel(overall_level)

    for handler in logger.handlers:
        # FileHandler is an instance of StreamHandler so need to check FileHandler first.
        if isinstance(handler, logging.FileHandler):
            handler.setLevel(file_level)
        elif isinstance(handler, logging.StreamHandler):
            handler.setLevel(stream_level)



from typing import Union, Dict, List, AnyStr, Any, TextIO, Literal

import tempfile
import xmltodict

from commons import *
from varvault import ResourceNotFoundError

vault_file_new = f"{DIR}/new-vault.xml"
existing_vault = f"{DIR}/existing-vault.xml"


class XmlResource(varvault.BaseResource):

    KEY = "VAULT"

    def __init__(self, path: str, mode: Literal["r", "w", "a", "r+", "w+", "a+"] = "r"):
        self.file_io = None
        super(XmlResource, self).__init__(path, mode)

    @property
    def resource(self) -> TextIO:
        return self.file_io

    def create(self) -> None:
        """Creates the resource self.file_io for this handler which we'll use to read and write to."""
        path = self.path
        assert path, "Path is not defined"
        dirname = os.path.dirname(path)

        create_dir = lambda: os.makedirs(dirname, exist_ok=True) if dirname else None
        write = lambda: self.do_write({})

        if self.mode_properties.create and not self.mode_properties.load:
            create_dir()
            write()

        elif self.mode_properties.load and self.mode_properties.create:
            if not self.exists():
                create_dir()
                write()

        elif self.mode_properties.load and not self.mode_properties.create:
            try:
                self.do_read()
            except Exception as e:
                if not self.mode_properties.live_update:
                    raise ResourceNotFoundError(f"Unable to read from resource at {path} (mode is {self.mode})", self) from e
                else:
                    return
        else:
            raise NotImplementedError(f"Mode {self.mode} is not valid ({self.mode_properties})")

        self.file_io = open(self.path, "r+")
        self.file_io.close()

    @property
    def state(self):
        import hashlib
        hash_md5 = hashlib.md5()
        with open(self.path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @property
    def path(self) -> Union[AnyStr, Any]:
        return self.raw_path

    def writable(self, obj: Dict) -> bool:
        obj = {self.KEY: obj}
        try:
            xmltodict.unparse(obj, pretty=True)
            return True
        except:
            return False

    def exists(self) -> bool:
        return os.path.exists(self.path)

    def do_write(self, vault: dict):
        # An XML file must have exactly one root. We assign the vault to the root self.KEY
        vault = {self.KEY: vault}
        xmltodict.unparse(vault, open(self.path, "w"), pretty=True)

    def do_read(self) -> Union[Dict, List]:
        vault = xmltodict.parse(open(self.path, "rb"))
        return vault[self.KEY]


class Keyring(varvault.Keyring):
    string_value = varvault.Key("string_value")
    int_value = varvault.Key("int_value")
    list_value = varvault.Key("list_value")


class TestXmlVault:
    @classmethod
    def setup_class(cls):
        tempfile.tempdir = "/tmp" if sys.platform == "darwin" or sys.platform == "linux" else tempfile.gettempdir()

    def setup_method(self):
        try:
            os.remove(vault_file_new)
        except:
            pass

    def test_create_xml_vault(self):
        vault = varvault.create(keyring=Keyring, resource=XmlResource(vault_file_new, mode="w"))

        @vault.manual(output=(Keyring.string_value, Keyring.int_value, Keyring.list_value))
        def _set():
            return "valid", 1, [1, 2, 3, 4, 5]

        _set()

        data_in_file = xmltodict.parse(open(vault_file_new, "rb"))[XmlResource.KEY]
        logger.info(data_in_file)
        assert Keyring.string_value in data_in_file and data_in_file[Keyring.string_value] == "valid"
        # XML data is actually just strings. But we still want to support using XML files, but it will take some manual labour to get it working correctly.
        assert Keyring.int_value in data_in_file and data_in_file[Keyring.int_value] == "1"
        assert Keyring.list_value in data_in_file

    def test_read_from_xml_vault(self):
        vault = varvault.create(keyring=Keyring, resource=XmlResource(existing_vault, "r"))

        @vault.manual(input=(Keyring.string_value, Keyring.int_value))
        def _get(string_value=None, int_value=None):
            assert string_value == "valid"
            assert int_value == "1"

        _get()




__version__ = "6.0.2"

import os
import json

from typing import Dict, TextIO, AnyStr, Literal, Union

from .resource import ResourceModes
from .resource import BaseResource
from .resource import ResourceNotFoundError

from .keyring import Key
from .keyring import Keyring

from .validator import validator
from .validator import modifier
from .validator import ValidatorException
from .validator import ModifierException

from .minivault import MiniVault

from .vault import VarVault

from .flags import Flags

from .factory import create

from .vaultstructs import VaultStructDictBase
from .vaultstructs import VaultStructListBase
from .vaultstructs import VaultStructFloatBase
from .vaultstructs import VaultStructIntBase
from .vaultstructs import VaultStructStringBase

from .utils import AssignedByVault
from .utils import concurrent_execution
from .utils import assert_and_raise


def clear_logs():
    import glob
    import tempfile

    path = os.path.join(tempfile.tempdir, "varvault-logs", "*.log")
    files = glob.glob(path)
    for f in files:
        os.remove(f)


class JsonResource(BaseResource):

    def __init__(self, path: AnyStr, mode: Union[Literal["r", "w", "a", "r+", "w+", "a+"], ResourceModes] = "r"):
        f"""
        Creates the JsonResource object.
        :param path: This should be the path to the JSON file that the vault should be using.
        :param mode: Sets the mode of the resource. The mode can be one of the following: 'r', 'w', 'a', 'r+', 'w+', 'a+'.
        r: Read from existing resource (default)
        w: Create new resource and ignore existing resource and write to it
        a: Create a new resource if none exist, otherwise read from and write to existing resource
        r+: Read from existing resource and perform live-update
        w+: Create new resource and ignore existing resource and write to it, and perform live-update
        a+: Create a new resource if none exist, otherwise read from and write to existing resource, and perform live-update
        """
        super(JsonResource, self).__init__(path, mode)
        self.file_io = None

    @property
    def state(self):
        """Returns the state of the vault, which is the state of the JSON file"""
        import hashlib
        hash_md5 = hashlib.md5()
        with open(self.path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @property
    def resource(self) -> TextIO:
        """Returns the file resource object for this handler."""
        return self.file_io

    def create(self) -> None:
        """Creates the resource self.file_io for this handler which we'll use to read and write to."""
        path = self.path
        assert path, "Path is not defined"
        dirname = os.path.dirname(path)

        create_dir = lambda: os.makedirs(dirname, exist_ok=True) if dirname else None
        write = lambda: self.do_write({})

        if self.mode_properties.create and not self.mode_properties.load:
            create_dir()
            write()

        elif self.mode_properties.load and self.mode_properties.create:
            if not self.exists():
                create_dir()
                write()
        else:
            assert_and_raise(self.mode_properties.load and not self.mode_properties.create, NotImplementedError(f"Mode {self.mode} is not valid ({self.mode_properties})"))
            try:
                self.do_read()
            except Exception as e:
                if not self.mode_properties.live_update:
                    raise ResourceNotFoundError(f"Unable to read from resource at {path} (mode is {self.mode})", self) from e
                else:
                    return

        self.file_io = open(self.path, "r+")
        self.file_io.close()

    @property
    def backup(self):
        """Returns the backup path for this handler"""
        return self.raw_path + ".bak"

    @property
    def path(self) -> AnyStr:
        """Returns the path to the vault-file as a JSON file."""
        return self.raw_path if os.path.exists(self.raw_path) else self.backup

    def writable(self, obj: Dict) -> bool:
        f"""Checks if a key-value pair in a dict can be written to a file by attempting to serialize it by using {json.dumps}"""
        try:
            json.dumps(obj)
            return True
        except (TypeError, OverflowError) as e:
            return False

    def exists(self) -> bool:
        """Returns a bool that determines if the JSON file exists by expanding user and potential vars"""
        return os.path.exists(self.path)

    def do_write(self, vault: dict) -> None:
        """Writes the vault to the JSON file"""
        with open(self.backup, "w") as f:
            json.dump(vault, f, indent=2)
        os.rename(self.backup, self.raw_path)

    def do_read(self) -> Dict:
        """Reads the vault from the JSON file"""
        return json.load(open(self.path))



import re
import inspect
import warnings
import functools

from typing import Callable

from .utils import assert_and_raise

_VALIDATOR = "_validator"
_MODIFIER = "_modifier"


class ValidatorException(Exception):
    pass


class ModifierException(Exception):
    pass


def validator(function_asserts: bool = None, function_returns_bool: bool = None, skip_source_assertions: bool = False) -> Callable:
    f"""
    Decorator that is used to register a function as a validator-function used for varvault-keys.
    A validator-function assigned to a key will be executed based on how it's registered.
    
    If no argument is passed to the decorator, {function_asserts} will be set to {True}.

    :param function_asserts: Used to tell varvault that the validator-function will do the validation by performing an assert
    :param function_returns_bool: Used to tell varvault that the validator-function will return a boolean that varvault will assert to be True.
    :param skip_source_assertions: Used to tell this decorator to skip source-assertions that verify that assert or return is done in the validator-function.  
    """
    assert isinstance(function_asserts, bool) or function_asserts is None, f"'function_asserts' must be a bool, or {None}"
    assert isinstance(function_returns_bool, bool) or function_returns_bool is None, f"'function_returns_bool' must be a bool, or {None}"
    if function_asserts and function_returns_bool:
        raise SyntaxError("You've set both 'function_asserts' and 'function_returns_bool' to True; 'function_asserts' takes precedence. Is this what you really want?")
    if not function_asserts and not function_returns_bool:
        function_asserts = True

    def wrap_outer(func: Callable) -> Callable:

        assert callable(func), f"Object {func} is not a callable"
        argspecs = inspect.getfullargspec(func)

        assert len(argspecs.args) == 1, f"You've defined more arguments (or no arguments) for the validator function than you are allowed to. There must be exactly one argument defined in the signature."

        source = inspect.getsource(func)
        if function_asserts:
            assert re.findall(re.compile(r".*assert .*"), source) or skip_source_assertions, "No assert-statement found in the functions source. That doesn't seem right if you are planning to do an assert in the function. Set 'skip_source_assertions=True' to skip it"
        elif function_returns_bool:
            if argspecs.annotations.get("return") != bool:
                warnings.warn("If you define the validator-function to say that it returns a bool, you should really also annotate the function to say it returns a bool.", SyntaxWarning)
            assert re.findall(re.compile(r".*return .*"), source) or skip_source_assertions, "No return-statement found in the functions source. That doesn't seem right if you are planning to return a bool that varvault is meant to assert. Set 'skip_source_assertions=True' to skip"

        setattr(func, _VALIDATOR, True)

        @functools.wraps(func)
        def wrap_inner(keyvalue: object, key):
            if function_asserts:
                try:
                    func(keyvalue)
                except AssertionError as e:
                    raise ValidatorException(f"{e} for key '{key}'") from e
            elif function_returns_bool:
                ret = func(keyvalue)
                assert_and_raise(isinstance(ret, bool), ValidatorException(f"The return value is of type {type(ret)}, not of type {bool}, and you have said "
                                                                           f"that the function will return a bool through 'function_returns_bool' being set to {True}"))
                assert_and_raise(ret, ValidatorException(f"Validator-function {func.__name__} returned {False} for key '{key}' with value: {keyvalue}"))
        return wrap_inner
    return wrap_outer


def modifier() -> Callable:
    f"""
    Decorator that is used to register a function as a modifier-function used for varvault-keys.
    A modifier-function will change the value of object assigned to the key before it's written to the vault. A good example of this is 
    to, for example, change the path of a file to an absolute path, or expand user/vars in the path of the file. 
    It can also be used to strip a sequence of characters from a string, or to change the case of a string.
    """

    def wrap_outer(func: Callable) -> Callable:

        assert callable(func), f"Object {func} is not a callable"
        argspecs = inspect.getfullargspec(func)

        assert len(argspecs.args) == 1, f"You've defined more arguments (or no arguments) for the validator function than you are allowed to. There must be exactly one argument defined in the signature."
        source = inspect.getsource(func)
        assert re.findall(re.compile(r".*return .*"), source), "No return-statement found in the functions source. That doesn't seem right if you are planning to modify the value of the object."
        from varvault import Key
        setattr(func, _MODIFIER, True)

        @functools.wraps(func)
        def wrap_inner(keyvalue: object, key: Key):
            try:
                return func(keyvalue)
            except Exception as e:
                raise ModifierException(f"{e} for key '{key}'") from e
        return wrap_inner
    return wrap_outer



import json

from commons import *


vault_file_new = f"{DIR}/new-vault.json"


class VaultStructDict(varvault.VaultStructDictBase):
    def __init__(self, value_1: str, value_2: int, **kwargs):
        super(VaultStructDict, self).__init__(**kwargs)
        assert isinstance(value_1, str)
        assert isinstance(value_2, int)

        self.value_1 = value_1
        self.value_2 = value_2

    def internal_function(self):
        pass

    @classmethod
    def create(cls, vault_key, vault_value):
        obj = VaultStructDict(**vault_value)
        return obj


class VaultStructList(varvault.VaultStructListBase):
    def __init__(self, value_1: str, value_2: int, *args):
        super(VaultStructList, self).__init__(*args)
        assert isinstance(value_1, str)
        assert isinstance(value_2, int)

        self.value_1 = value_1
        self.value_2 = value_2

        self.extend([value_1, value_2])

    def internal_function(self):
        pass

    @classmethod
    def create(cls, vault_key, vault_value):
        obj = VaultStructList(*vault_value)
        return obj


class VaultStructString(varvault.VaultStructStringBase):
    def __new__(cls, string_value, extra_value, *args, **kwargs):
        assert isinstance(string_value, str)
        obj = super().__new__(cls, string_value)
        obj.string_value = string_value
        obj.extra_value = extra_value
        return obj

    def internal_function(self):
        pass

    @classmethod
    def create(cls, vault_key, vault_value):
        obj = VaultStructString(string_value=vault_value, extra_value="extra_value-cannot-possibly-be-saved-to-a-string")
        return obj


class VaultStructFloat(varvault.VaultStructFloatBase):
    def __new__(cls, float_value, extra_value, *args, **kwargs):
        assert isinstance(float_value, float)
        obj = super().__new__(cls, float_value)
        obj.float_value = float_value
        obj.extra_value = extra_value
        return obj

    def internal_function(self):
        pass

    @classmethod
    def create(cls, vault_key, vault_value):
        obj = VaultStructFloat(float_value=vault_value, extra_value="extra_value-cannot-possibly-be-saved-to-a-float")
        return obj


class VaultStructInt(varvault.VaultStructIntBase):
    def __new__(cls, int_value, extra_value, *args, **kwargs):
        assert isinstance(int_value, int)
        obj = super().__new__(cls, int_value)
        obj.int_value = int_value
        obj.extra_value = extra_value
        return obj

    def internal_function(self):
        pass

    @classmethod
    def create(cls, vault_key, vault_value):
        obj = VaultStructInt(int_value=vault_value, extra_value="extra_value-cannot-possibly-be-saved-to-an-int")
        return obj


class VaultStructDictInvalid(varvault.VaultStructDictBase):
    # This struct should fail to be loaded from a file since "create" is not implemented.

    def __init__(self, value_1: str, value_2: int, **kwargs):
        super(VaultStructDictInvalid, self).__init__(**kwargs)
        assert isinstance(value_1, str)
        assert isinstance(value_2, int)

        self.value_1 = value_1
        self.value_2 = value_2


class KeyringVaultStruct(varvault.Keyring):
    key_vault_struct_dict = varvault.Key("key_vault_struct_dict", valid_type=VaultStructDict)
    key_vault_struct_list = varvault.Key("key_vault_struct_list", valid_type=VaultStructList)
    key_vault_struct_string = varvault.Key("key_vault_struct_string", valid_type=VaultStructString)
    key_vault_struct_float = varvault.Key("key_vault_struct_float", valid_type=VaultStructFloat)
    key_vault_struct_int = varvault.Key("key_vault_struct_int", valid_type=VaultStructInt)
    key_vault_struct_dict_invalid = varvault.Key("key_vault_struct_dict_invalid", valid_type=VaultStructDictInvalid)


class TestVaultStructs:
    def setup_method(self):
        try:
            os.remove(vault_file_new)
        except:
            pass

    def test_vault_struct_dict(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_dict)
        def _set():
            return VaultStructDict("v1", 1)

        _set()
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_dict), VaultStructDict)
        logger.info(vault.get(KeyringVaultStruct.key_vault_struct_dict))
        vault_struct_dict = vault.get(KeyringVaultStruct.key_vault_struct_dict)
        vault_struct_dict["value_1"] = "v2"
        vault_struct_dict["value_2"] = 2
        vault.insert(KeyringVaultStruct.key_vault_struct_dict, vault_struct_dict, varvault.Flags.permit_modifications)

        from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert isinstance(from_vault.get(KeyringVaultStruct.key_vault_struct_dict), VaultStructDict)
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_dict), "internal_function")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_dict), "value_1")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_dict), "value_2")

    def test_vault_struct_list(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_list)
        def _set():
            return VaultStructList("v1", 1)

        _set()
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_list), VaultStructList)
        logger.info(vault.get(KeyringVaultStruct.key_vault_struct_list))

        from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert isinstance(from_vault.get(KeyringVaultStruct.key_vault_struct_list), VaultStructList)
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_list), "internal_function")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_list), "value_1")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_list), "value_2")

    def test_vault_struct_string(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_string)
        def _set():
            return VaultStructString("string-value", "extra-value-here")

        _set()
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_string), VaultStructString)
        logger.info(vault.get(KeyringVaultStruct.key_vault_struct_string))

        from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert isinstance(from_vault.get(KeyringVaultStruct.key_vault_struct_string), VaultStructString)
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_string), "internal_function")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_string), "string_value")

    def test_vault_struct_float(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_float)
        def _set():
            return VaultStructFloat(3.14, "extra-value-here")

        _set()
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_float), VaultStructFloat)
        logger.info(vault.get(KeyringVaultStruct.key_vault_struct_float))

        from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert isinstance(from_vault.get(KeyringVaultStruct.key_vault_struct_float), VaultStructFloat)
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_float), "internal_function")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_float), "float_value")

    def test_vault_struct_int(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_int)
        def _set():
            return VaultStructInt(1, "extra-value-here")
        _set()

        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_int), VaultStructInt)
        logger.info(vault.get(KeyringVaultStruct.key_vault_struct_int))

        from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert isinstance(from_vault.get(KeyringVaultStruct.key_vault_struct_int), VaultStructInt)
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_int), "internal_function")
        assert hasattr(from_vault.get(KeyringVaultStruct.key_vault_struct_int), "int_value")

    def test_invalid_vault_struct(self):
        vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault.manual(output=KeyringVaultStruct.key_vault_struct_dict_invalid)
        def _set():
            return VaultStructDictInvalid("v1", 1)

        _set()
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_dict_invalid), VaultStructDictInvalid)
        try:
            from_vault = varvault.create(keyring=KeyringVaultStruct, resource=varvault.JsonResource(vault_file_new, mode="r"))
            assert False, "Managed to create a vault with a VaultStruct that hasn't been implemented correctly"
        except NotImplementedError as e:
            logger.info(f"Expected error {e}: failed to load vault with key as invalid VaultStruct")

    def test_with_live_update(self):
        resource = varvault.JsonResource(vault_file_new, mode="w+")
        vault = varvault.create(keyring=KeyringVaultStruct, resource=resource)
        extra = {KeyringVaultStruct.key_vault_struct_int: VaultStructInt(1, 1)}
        state = resource.state
        logger.info(f"After create: {state}")
        mv = varvault.MiniVault({KeyringVaultStruct.key_vault_struct_dict: VaultStructDict("v1", 1),
                                 KeyringVaultStruct.key_vault_struct_list: VaultStructList("v1", 1),
                                 KeyringVaultStruct.key_vault_struct_string: VaultStructString("v1", 1),
                                 KeyringVaultStruct.key_vault_struct_float: VaultStructFloat(3.14, 1)})
        vault.insert_minivault(mv)
        logger.info(f"After insert: {resource.state}")
        assert state != resource.state, "State of the file appears to have not changed"
        state = resource.state
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_dict), VaultStructDict), type(vault.get(KeyringVaultStruct.key_vault_struct_dict))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_list), VaultStructList), type(vault.get(KeyringVaultStruct.key_vault_struct_list))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_string), VaultStructString), type(vault.get(KeyringVaultStruct.key_vault_struct_string))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_float), VaultStructFloat), type(vault.get(KeyringVaultStruct.key_vault_struct_float))
        assert state == resource.state, "State of the file has changed when it should have stayed the same. For some reason the vault updated the file unexpectedly"
        state = resource.state
        logger.info(f"After read: {resource.state}")

        d = json.load(open(vault_file_new))
        d.update(extra)
        json.dump(d, open(vault_file_new, "w"))

        logger.info(f"After update: {resource.state}")
        assert state != resource.state, "State of the file appears to have not changed, but we just added an extra key to the file"

        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_dict), VaultStructDict), type(vault.get(KeyringVaultStruct.key_vault_struct_dict))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_list), VaultStructList), type(vault.get(KeyringVaultStruct.key_vault_struct_list))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_string), VaultStructString), type(vault.get(KeyringVaultStruct.key_vault_struct_string))
        assert isinstance(vault.get(KeyringVaultStruct.key_vault_struct_float), VaultStructFloat), type(vault.get(KeyringVaultStruct.key_vault_struct_float))

        logger.info(f"After second read: {resource.state}")



import tempfile

import varvault.logger
from commons import *


vault_file_new = f"{DIR}/new-vault.json"
vault_file_new_secondary = f"{DIR}/new-vault-secondary.json"


class TestLogging:
    @classmethod
    def setup_class(cls):
        tempfile.tempdir = "/tmp" if sys.platform == "darwin" or sys.platform == "linux" else tempfile.gettempdir()

    def setup_method(self):
        try:
            temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault-vault-stream.log")
            os.remove(temp_log_file)
        except:
            pass

        try:
            vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
            os.remove(vault_log_file)
        except:
            pass

        try:
            os.remove(vault_file_new)
        except:
            pass
        try:
            os.remove(vault_file_new_secondary)
        except:
            pass

    def test_silent(self):
        temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault-vault-stream.log")
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        vault_new = varvault.create(varvault.Flags.silent, varvault.Flags.remove_existing_log_file, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        vault_new.logger.addHandler(logging.StreamHandler(open(temp_log_file, "w")))

        @vault_new.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()
        assert len(open(temp_log_file).readlines()) <= 2, f"There appears to be more lines in the log file than what there should be. " \
                                                          f"There should only be 2 at most. {varvault.Flags.silent} appears to not function correctly"
        assert len(open(vault_log_file).readlines()) <= 2, f"There appears to be more lines in the log file than what there should be. " \
                                                           f"There should only be 2 at most. {varvault.Flags.silent} appears to not function correctly. " \
                                                           f"Contents: \n{open(vault_log_file).readlines()}"

    def test_debug(self):
        temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault-vault-stream.log")
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        vault_new = varvault.create(varvault.Flags.debug, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        # Create and set a file to act as a StreamHandler for the logger object in varvault.
        # This way, we can easily capture stdout to a file and assert that the output is the expected
        vault_new.logger.addHandler(logging.StreamHandler(open(temp_log_file, "w")))

        @vault_new.manual(output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()

        assert len(open(temp_log_file).readlines()) >= 10, f"There appears to be fewer lines in the log file than what there should be. " \
                                                           f"There should only be 12 at least. {varvault.Flags.debug} appears to not function correctly"
        assert len(open(vault_log_file).readlines()) >= 12, f"There appears to be fewer lines in the log file than what there should be. " \
                                                            f"There should only be 12 at least. {varvault.Flags.debug} appears to not function correctly"

        @vault_new.manual(input=Keyring.key_valid_type_is_str)
        def _use(key_valid_type_is_str: str = varvault.AssignedByVault):
            return key_valid_type_is_str

        assert _use() == "valid", "The value returned by the function is not the same as the value that was set"

    def test_silent_and_debug(self):
        temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault-vault-stream.log")
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        vault_new = varvault.create(varvault.Flags.debug, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        # Create and set a file to act as a StreamHandler for the logger object in varvault.
        # This way, we can easily capture stdout to a file and assert that the output is the expected
        vault_new.logger.addHandler(logging.StreamHandler(open(temp_log_file, "w")))

        @vault_new.manual(varvault.Flags.silent, output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()

        assert len(open(temp_log_file).readlines()) == 0, f"There appears to be more lines in the log file than what there should be. " \
                                                          f"There should be 0 at most. {varvault.Flags.debug} with {varvault.Flags.silent} appears to not function correctly"
        assert len(open(vault_log_file).readlines()) == 12, f"There appears to be fewer lines in the log file than what there should be. " \
                                                            f"There should be 12 at most. {varvault.Flags.debug} with {varvault.Flags.silent} appears to not function correctly"

    def test_disable_logger(self):
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        try:
            os.unlink(vault_log_file)
        except OSError:
            pass
        assert not os.path.exists(vault_log_file), f"{vault_log_file} still exists, weird"

        vault_new = varvault.create(varvault.Flags.disable_logger, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        assert vault_new.logger is None, "logger object is not None; it should be"
        assert not os.path.exists(vault_log_file), f"{vault_log_file} exists after creating the vault when saying there shouldn't be a logger object"

        @vault_new.manual(varvault.Flags.silent, output=Keyring.key_valid_type_is_str)
        def _set():
            return "valid"
        _set()
        assert not os.path.exists(vault_log_file), f"{vault_log_file} exists after using the vault. How?!"

    def test_remove_existing_log_file(self):
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        vault_new = varvault.create(varvault.Flags.debug, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))

        @vault_new.manual(varvault.Flags.silent, output=Keyring.key_valid_type_is_str)
        def _doset():
            return "valid"
        _doset()
        with open(vault_log_file) as f1:
            vault_log_file_lines = f1.readlines()
            assert len(vault_log_file_lines) == 12, f"There should be exactly 12 lines in the log-file:\n{''.join(vault_log_file_lines)}"
        vault_from = varvault.create(varvault.Flags.remove_existing_log_file, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="r"))
        assert Keyring.key_valid_type_is_str in vault_from
        with open(vault_log_file) as f2:
            vault_log_file_lines = f2.readlines()
            assert len(vault_log_file_lines) == 2, f"There should be exactly 2 lines in the logfile. It seems the log-file wasn't removed when the new vault was created from the existing vault: \n{''.join(vault_log_file_lines)}"

    def test_specific_logger(self):
        old_handlers = logger.handlers.copy()
        temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "pytest-stream.log")
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "pytest-file.log")
        try:
            os.unlink(temp_log_file)
        except OSError:
            pass
        try:
            os.unlink(vault_log_file)
        except OSError:
            pass

        try:
            logger.handlers.clear()
            logger.addHandler(logging.StreamHandler(open(temp_log_file, "w")))
            logger.addHandler(logging.FileHandler(filename=vault_log_file))

            vault_new = varvault.create(varvault.Flags.debug, varvault.Flags.remove_existing_log_file, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"), logger=logger)
            assert vault_new.logger.name == "pytest"  # The logger used for pytest here is called pytest

            @vault_new.manual(varvault.Flags.silent, output=Keyring.key_valid_type_is_str)
            def _set():
                return "valid"
            _set()
            temp_log_file_lines = open(temp_log_file).readlines()
            assert len(temp_log_file_lines) == 1, f"There appears to be more lines in the log file than what there should be. There should be 1 at most. \n{''.join(temp_log_file_lines)}"
            vault_log_file_lines = open(vault_log_file).readlines()
            assert len(vault_log_file_lines) == 11, f"There appears to be fewer lines in the log file than what there should be. There should be 11 at most. \n{''.join(vault_log_file_lines)}"

        finally:
            logger.handlers.clear()
            logger.handlers = old_handlers

    def test_no_error_logging(self):
        temp_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault-vault-stream.log")
        vault_log_file = os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")
        vault_new = varvault.create(varvault.Flags.debug, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        # Create and set a file to act as a StreamHandler for the logger object in varvault.
        # This way, we can easily capture stdout to a file and assert that the output is the expected
        vault_new.logger.addHandler(logging.StreamHandler(open(temp_log_file, "w")))

        @vault_new.manual(varvault.Flags.no_error_logging, output=Keyring.key_valid_type_is_str)
        def _set():
            raise Exception("Failing deliberately")

        try:
            _set()
        except:
            pass

        assert len(open(temp_log_file).readlines()) == 3, f"There appears to be more lines in the log file than what there should be. " \
                                                          f"There should be 3 at most. It appears that {varvault.Flags.no_error_logging} doesn't work properly"
        assert len(open(vault_log_file).readlines()) == 5, f"There appears to be fewer lines in the log file than what there should be. " \
                                                           f"There should be 5 at most. It appears that {varvault.Flags.no_error_logging} doesn't work properly"

    def test_clear_logs(self):
        vault = varvault.create(varvault.Flags.debug, keyring=Keyring, resource=varvault.JsonResource(vault_file_new, mode="w"))
        varvault.clear_logs()
        assert not os.path.exists(os.path.join(tempfile.gettempdir(), "varvault-logs", "varvault.log")), f"The log file still exists after calling {varvault.clear_logs.__name__}"


import glob
import pathlib
import re

from setuptools import setup

import varvault

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = pathlib.Path(f"{HERE}/README.md").read_text()

# Version handling
# A good way to tell if we are backwards compatible is to run the test suite and if the tests pass without requiring any changes, we can pretty safely assume we are backwards compatible.
MAJOR = 6  # Change this if the previous MAJOR is incompatible with this build. Set MINOR and PATCH to 0
MINOR = 0  # Change this if the functionality has changed, but we are still backwards compatible with previous MINOR versions. Set PATCH to 0
PATCH = 2  # Change this is if we are fixing a bug that doesn't change the functionality. If a bug-fix has caused functionality to be changed, see MINOR instead
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"


assert varvault.__version__ == VERSION, f"Version mismatch: {varvault.__version__} != {VERSION}"


def find_todos_for_version_in_code():
    """
    Checks for TODOs/FIXMEs in the code linked to the current version
    :return: A list of lines in the code that are marked with FIX ME or TO DO linked to the current version.
    """
    """This pattern should match these lines (given VERSION is 2.0.0): 
    some_code = "code"  # FIXME: Remove this in 2.0.0
    # TODO: Remove this in 2.0.0
    # fixme: Remove this in 2.0.0
    # todo: Remove this in 2.0.0
    # fix me: Remove this in 2.0.0
    # to do: Remove this in 2.0.0
    """
    pattern = re.compile(rf".*(#.*(?:[Ff][Ii][Xx].*[Mm][Ee]|[Tt][Oo].*[Dd][Oo]).*{MAJOR}\.{MINOR}\.{PATCH}.*)")
    files = glob.glob("varvault/*.py")
    matches = list()
    for file in files:
        with open(file) as f:
            for i, line in enumerate(f.readlines()):
                match = re.match(pattern, line)
                if match:
                    matches.append((file, i, match.group(1)))
    return matches


todos_in_code = find_todos_for_version_in_code()
assert len(todos_in_code) == 0, f"There are FIXMEs/TODOs in the code that appear to be linked to this version; You should fix these before making the release: {todos_in_code}"


# This call to setup() does all the work
setup(
    name="varvault",
    version=VERSION,
    description="A package that sets up a key-value vault to store and access variables in a global context.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/data-ductus/varvault",
    author="Chloe Holst",
    author_email="chloe.holst@dataductus.se",
    license="Apache 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8"
    ],
    packages=["varvault"],
    include_package_data=True,
)
